import os
import subprocess
import sys
import shutil
from tempfile import mkdtemp
import numpy as np
from openEMS import openEMS
from CSXCAD.CSXCAD import ContinuousStructure
from pyems.boundary import BoundaryConditions


class Simulation:
    """
    OpenEMS simulation.  This is the main entry-point for
    creating/running a simulation.
    """

    def __init__(
        self,
        freq: np.array,
        unit: float = 1,
        boundary_conditions: BoundaryConditions = BoundaryConditions(
            (("PML_8", "PML_8"), ("PML_8", "PML_8"), ("PML_8", "PML_8")),
        ),
        reference_frequency: float = None,
        end_criteria: float = 1e-5,
        timestep_factor: float = 1,
        sim_dir: str = "sim",
        calc_only: bool = False,
    ):
        """
        :param freq: An ordered (ascending) numpy array or list of
            frequency values to simulate.  All signal excitations will
            be automatically set based on this value, where the middle
            frequency value determines the center frequency of the
            gaussian excitation and the end frequency values are the
            -20dB values.  Additionally, the simulation timestep and
            post-processing frequency results are both dependent on
            this argument.  A larger number of values will increase
            the simulation time but will also increase the frequency
            resolution of output data.
        :param unit: Length dimension unit to use for all simulation
            distances.  Defaults to 1, which is meters.  For instance,
            1e-3 would be mm.
        :param boundary_conditions: The OpenEMS simulation boundary
            conditions.  Corresponds to ((xmin, xmax), (ymin, ymax),
            (zmin, zmax)).  See the OpenEMS documentation for details.
        :param reference_frequency: Certain dielectric properties are
            frequency dependent.  OpenEMS can only use a single value
            for these parameters during a simulation.  This parameter
            sets the frequency to use.  If the default None is given,
            the center frequency from the frequency array is used.
        :param end_criteria: FDTD termination energy.
        :param timestep_factor: Can reduce the timestep for added
            stability.  This must be between 0 (exclusive) and 1
            (inclusive).
        :param sim_dir: Directory where simulation results are stored.
            If you pass None, a temporary directory will be used.
            It's generally not recommended to use a temporary
            directory since it makes it more difficult to abort the
            simulation and to avoid rerunning unnecessary parts of the
            simulation.
        :param calc_only: Will not clear a previous simulation and run
            a new one.  This looks for existing simulation results in
            `sim_dir` and calculates results for it.  Unfortunately,
            Simulation still needs to be told about all ports and
            simulation parameters in order to interpret the simulation
            results.
        """
        self._freq = np.array(freq)
        self._unit = unit
        self._csx = ContinuousStructure()
        self._csx.GetGrid().SetDeltaUnit(self._unit)
        self._boundary_conditions = boundary_conditions
        if reference_frequency is None:
            self._reference_frequency = self.center_frequency()
        else:
            self._reference_frequency = reference_frequency
        self._end_criteria = end_criteria
        self._calc_only = calc_only
        if sim_dir is None:
            self._sim_dir = mkdtemp()
        else:
            sim_dir = os.path.abspath(sim_dir)
            if not calc_only:
                if os.path.exists(sim_dir):
                    shutil.rmtree(sim_dir)
                os.mkdir(sim_dir)
            self._sim_dir = sim_dir
        if timestep_factor <= 0 or timestep_factor > 1:
            raise ValueError(
                "timestep_factor must be between 0 (exclusive) "
                "and 1 (inclusive)."
            )
        self._fdtd = openEMS(
            EndCriteria=self._end_criteria, TimeStepFactor=timestep_factor
        )
        self._fdtd.SetGaussExcite(
            self.center_frequency(), self.half_bandwidth()
        )
        self._fdtd.SetBoundaryCond(self.boundary_conditions.as_list())
        self._fdtd.SetCSX(self._csx)
        self._ports = []
        self._field_dumps = []
        self._mesh = None
        self._csx_path = None
        self._nf2ff = None

    @property
    def freq(self) -> np.array:
        """
        """
        return self._freq

    @property
    def unit(self) -> float:
        """
        """
        return self._unit

    @property
    def csx(self) -> ContinuousStructure:
        """
        """
        return self._csx

    @property
    def fdtd(self) -> openEMS:
        """
        """
        return self._fdtd

    @property
    def boundary_conditions(self) -> BoundaryConditions:
        """
        """
        return self._boundary_conditions

    @property
    def ports(self):
        """
        """
        return self._ports

    @property
    def mesh(self):
        """
        """
        return self._mesh

    @property
    def sim_dir(self):
        """
        """
        return self._sim_dir

    @property
    def nf2ff(self):
        """
        """
        return self._nf2ff

    def run(self, csx: bool = True, debug_pec: bool = False) -> None:
        """
        """
        if csx:
            self.view_csx(prompt=True)
        if not self._calc_only:
            if debug_pec:
                self.fdtd.Run(self.sim_dir, setup_only=True, debug_pec=True)
            else:
                self.fdtd.Run(self.sim_dir, cleanup=False)
                self._calc_ports()

    def view_csx(self, prompt: bool = False) -> None:
        """
        View the CSX network.

        :param prompt: Prompt user whether to continue simulation.
        """
        subprocess.run(["AppCSXCAD", self._csx_path])
        if prompt:
            self._prompt_terminate()

    def view_field(self, index: int = 0) -> None:
        """
        View the field dump corresponding to the given index.
        """
        if index > len(self._field_dumps) - 1:
            raise ValueError("Invalid field dump index provided.")
        self._field_dumps[index].view()

    def _prompt_terminate(self) -> None:
        """
        """
        ans = input("Continue simulation (y/n)? ")
        ans = ans.lower()
        if ans == "n":
            print("Terminating simulation.")
            sys.exit(0)
        elif ans != "y":
            print("Please answer y/n.")
            self._prompt_terminate()

    def post_mesh(self):
        """
        Adjust the simulation for the generated mesh.
        """
        self._align_ports_to_mesh()
        self.save_csx()
        self._mesh_errors()

    def _mesh_errors(self) -> None:
        """
        """
        for port in self.ports:
            if port.pml_overlap():
                raise RuntimeError(
                    "Probe or feed overlaps PML. Please fix your simulation. "
                    "CSX file has been saved so you can view the overlap."
                )

    def save_csx(self, path: str = None):
        """
        """
        if path is None:
            path = self.sim_dir + "/" + self._file_name() + ".csx"
        self.csx.Write2XML(path)
        self._csx_path = path

    def _file_name(self):
        """
        """
        return os.path.splitext(os.path.basename(sys.argv[0]))[0]

    def _calc_ports(self):
        """
        """
        [port.calc() for port in self.ports]

    def s_param(self, i: int, j: int, dB: bool = True) -> np.array:
        """
        Calculate the S-parameter, S_{ij}.

        :param i: First subscript of S.  Must be in the range [1,
                  num_ports]
        :param j: Second subscript of S.  Must be in the range [1,
                  num_ports]
        :param dB: Return the s-parameter as a decibel value.
        """
        num_ports = self._num_ports()
        if i > num_ports or j > num_ports or i < 1 or j < 1:
            raise ValueError(
                "Invalid S-parameter requested. Ensure that i and j are in "
                "the proper range for the network."
            )

        i -= 1
        j -= 1
        s = (
            self.ports[i].reflected_voltage()
            / self.ports[j].incident_voltage()
        )
        s = np.abs(s)
        if not dB:
            return s
        else:
            return 20 * np.log10(s)

    def _num_ports(self) -> int:
        """
        """
        return len(self.ports)

    def center_frequency(self) -> float:
        """
        """
        idx = int(len(self.freq) / 2)
        return self.freq[idx]

    @property
    def reference_frequency(self) -> float:
        """
        """
        return self._reference_frequency

    def half_bandwidth(self) -> float:
        """
        """
        return self.freq[-1] - self.center_frequency()

    def max_frequency(self) -> float:
        """
        """
        return self.center_frequency() + self.half_bandwidth()

    def add_port(self, port) -> None:
        """
        """
        self._ports.append(port)
        self._order_ports()

    def add_field_dump(self, field_dump) -> None:
        """
        """
        self._field_dumps.append(field_dump)

    def _order_ports(self) -> None:
        """
        """
        self._ports.sort(key=lambda port: port.number)

    def _align_ports_to_mesh(self) -> None:
        """
        """
        if self.ports is not None:
            [port.snap_to_mesh(mesh=self.mesh) for port in self.ports]

    def register_mesh(self, mesh) -> None:
        """
        """
        self._mesh = mesh

    def register_nf2ff(self, nf2ff) -> None:
        """
        """
        self._nf2ff = nf2ff



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Matt Huszagh (huszaghmatt@gmail.com)
# Copyright (C) 2015,2016 Thorsten Liebig (Thorsten.Liebig@gmx.de)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
A port is an abstraction of an electrical port.  It provides an entry
point to an electrical network and performs the calculations needed to
determine S-parameters and other useful quantities.

This module overlaps significantly with OpenEMS's own implementation
of ports.  It reimplements many parts and, in other cases, delegates
to it for functionality.  The reason for reimpleenting this
functionality here is to allow better integration with automatic mesh
generation and to allow more flexible port creation.  OpenEMS
generally requires that a mesh be defined before the constituent
ports, which prevents mesh generation from being port-aware.  Port
awareness is necessary for proper mesh generation because ports
contain physical structures in addition to their non-physical
structures.  There are good reasons for generating ports based on an
existing mesh (e.g. voltage probes must be placed on mesh lines and
current probes must be placed between mesh lines, etc.), but this
module takes the position that we can always modify probe positions
after mesh generation.
"""

from typing import Tuple
from copy import deepcopy
from abc import ABC, abstractmethod
import numpy as np
from pyems.simulation import Simulation
from pyems.coordinate import Box3, Coordinate3, Axis
from pyems.mesh import Mesh
from pyems.probe import Probe
from pyems.utilities import array_index
from pyems.calc import wavenumber
from pyems.feed import Feed
from pyems.priority import priorities
from pyems.physical_constant import Z0


class Port(ABC):
    """
    Port base class.
    """

    def __init__(self, sim: Simulation, number: int, excite: bool = False):
        """
        :param sim: The Simulation to which this port is added.
        :param excite: Set to True if this port should generate an
            excitation.  The actual excitation type is set by the
            `Simulation` object that contains this port.
        """
        self._sim = sim
        self._number = number
        self.excite = excite
        self.feeds = []
        self.vprobes = []
        self.iprobes = []
        self._data_read = False
        self.v_inc = None
        self.v_ref = None
        self.i_inc = None
        self.i_ref = None
        self.p_inc = None
        self.p_ref = None
        self.z0 = None
        self.beta = None

        self.sim.add_port(self)

    @property
    def sim(self) -> Simulation:
        """
        """
        return self._sim

    @property
    def number(self) -> int:
        """
        Port number.
        """
        return self._number

    def snap_to_mesh(self, mesh: Mesh) -> None:
        """
        Generate the probes and feed and ensure they're located correctly
        in relation to the mesh.  You must call this in order to get
        correct simulation behavior.
        """
        self._set_probes(mesh=mesh)
        self._set_feed(mesh=mesh)

    def pml_overlap(self) -> bool:
        """
        Indicate if a probe or feed overlaps a PML boundary.

        :returns: True if there is overlap.
        """
        for vprobe in self.vprobes:
            if vprobe.pml_overlap():
                return True
        for iprobe in self.iprobes:
            if iprobe.pml_overlap():
                return True
        for feed in self.feeds:
            if feed.pml_overlap():
                return True
        return False

    def incident_voltage(self) -> np.array:
        """
        Get the incident voltage.  This can be used to calculate
        S-parameters.

        :returns: Array of the incident voltage magnitude for each
            frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        return self.v_inc

    def reflected_voltage(self) -> np.array:
        """
        Get the reflected voltage.  This can be used to calculate
        S-parameters.

        :returns: Array of the reflected voltage magnitude for each
            frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        return self.v_ref

    def incident_current(self) -> np.array:
        """
        Get the incident current.

        :returns: Array of the incident current magnitude for each
            frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        return self.i_inc

    def reflected_current(self) -> np.array:
        """
        Get the reflected current.

        :returns: Array of the reflected current magnitude for each
            frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        return self.i_ref

    def impedance(self, freq: float = None) -> np.array:
        """
        Get the characteristic impedance.

        :param freq: The frequency at which the impedance should be
            calculated.  If the the provided frequency doesn't
            correspond to a frequency bin, the nearest frequency bin
            will be selected.  If left as the default None, all
            frequency values will be given.

        :returns: Array of the characteristic impedance magnitude for
                  each frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        if freq is None:
            return self.z0

        idx = array_index(freq, self.sim.freq)
        return self.z0[idx]

    def incident_power(self) -> np.array:
        """
        Get the incident power.  This is generally useful for
        calculating efficiency.

        :returns: Array of the incident power magnitude for each
            frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        return self.p_inc

    def reflected_power(self) -> np.array:
        """
        Get the reflected power.

        :returns: Array of the reflected power magnitude for each
            frequency bin.
        """
        if not self._data_readp():
            raise RuntimeError("Must call calc() before retreiving values.")
        return self.p_ref

    def _calc_v_inc(self, v, i) -> None:
        """
        Calculate the incident voltage.

        See Pozar 4e section 4.3 (p.185) for derivation.

        :param v: Total voltage.
        :param i: Total current.
        """
        self.v_inc = (v + (self._ref_impedance * i)) / 2

    def _calc_v_ref(self, v, i) -> None:
        """
        Calculate the reflected voltage.

        See Pozar 4e section 4.3 (p.185) for derivation.

        :param v: Total voltage.
        :param i: Total current.
        """
        self.v_ref = (v - (self._ref_impedance * i)) / 2

    def _calc_i_inc(self, v, i) -> None:
        """
        Calculate the incident current.

        :param v: Total voltage.
        :param i: Total current.
        """
        self.i_inc = (i + (v / self._ref_impedance)) / 2

    def _calc_i_ref(self, v, i) -> None:
        """
        Calculate the reflected current.

        :param v: Total voltage.
        :param i: Total current.
        """
        self.i_ref = ((v / self._ref_impedance) - i) / 2

    def _calc_p_inc(self) -> None:
        """
        Calculate the port's incident power wave.
        """
        if self.v_inc is None or self.i_inc is None:
            raise RuntimeError(
                "Must calculate incoming and reflected "
                "voltages / current before power."
            )
        else:
            self.p_inc = (1 / 2) * np.real(self.v_inc * np.conj(self.i_inc))

    def _calc_p_ref(self) -> None:
        """
        Calculate the port's reflected power wave.
        """
        if self.v_ref is None or self.i_ref is None:
            raise RuntimeError(
                "Must calculate incoming and reflected "
                "voltages / current before power."
            )
        else:
            self.p_ref = (1 / 2) * np.real(self.v_ref * np.conj(self.i_ref))

    def _calc_beta(self, v, i, dv, di) -> None:
        """
        Calculate the transmission line propagation constant.

        Use tx line equations (see Pozar ch.2 for derivation):

        ..  math:: dV/dz = -(R+jwL)I

        ..  math:: dI/dz = -(G+jwC)V

        ..  math:: \gamma = \sqrt{(R+jwL)(G+jwC)}
        """
        self.beta = np.sqrt(-dv * di / (i * v))
        self.beta[np.real(self.beta) < 0] *= -1

    def _calc_z0(self, v, i, dv, di) -> None:
        """
        Calculate the transmission line characteristic impedance.

        Use tx line equations (see Pozar ch.2 for derivation):

        ..  math:: dV/dz = -(R+jwL)I

        ..  math:: dI/dz = -(G+jwC)V

        ..  math:: Z0 = \sqrt{(R+jwL)/(G+jwC)}
        """
        self.z0 = np.sqrt(v * dv / (i * di))
        if self._ref_impedance is None:
            self._ref_impedance = self.z0

    def _probe_vi(self) -> Tuple[np.array, np.array, np.array, np.array]:
        """
        Read probe voltage and current values.

        :returns: (v, i, dv, di)
        """
        [vprobe.read() for vprobe in self.vprobes]
        [iprobe.read() for iprobe in self.iprobes]
        prop_axis = self.propagation_axis().axis
        self._data_read = True
        v = self.vprobes[1].get_freq_data()[1]
        i = 0.5 * (
            self.iprobes[0].get_freq_data()[1]
            + self.iprobes[1].get_freq_data()[1]
        )
        dv = (
            self.vprobes[2].get_freq_data()[1]
            - self.vprobes[0].get_freq_data()[1]
        ) / (
            self.sim.unit
            * np.abs(
                self.vprobes[2].box.min_corner[prop_axis]
                - self.vprobes[0].box.min_corner[prop_axis]
            )
        )
        di = (
            self.iprobes[1].get_freq_data()[1]
            - self.iprobes[0].get_freq_data()[1]
        ) / (
            self.sim.unit
            * np.abs(
                self.iprobes[1].box.min_corner[prop_axis]
                - self.iprobes[0].box.min_corner[prop_axis]
            )
        )

        return (v, i, dv, di)

    def calc(self) -> None:
        """
        Calculate the characteristic impedance, propagation constant,
        and incident and reflected power.
        """
        v, i, dv, di = self._probe_vi()

        self._calc_beta(v, i, dv, di)
        self._calc_z0(v, i, dv, di)
        self._calc_v_inc(v, i)
        self._calc_v_ref(v, i)
        self._calc_i_inc(v, i)
        self._calc_i_ref(v, i)
        self._calc_p_inc()
        self._calc_p_ref()

    @abstractmethod
    def propagation_axis(self) -> Axis:
        """
        """
        pass

    def _data_readp(self) -> bool:
        """
        Return True if data has been read from simulation result.
        """
        return self._data_read


class MicrostripPort(Port):
    """
    Microstrip transmission line port.
    """

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        propagation_axis: Axis,
        excitation_axis: Axis,
        number: int,
        thickness: float,
        conductivity: float = 5.8e7,
        excite: bool = False,
        feed_impedance: float = None,
        feed_shift: float = 0.2,
        ref_impedance: float = None,
        measurement_shift: float = 0.5,
    ):
        """
        :param sim: Simulation to which microstrip port is added.
        :param box: 3D box specifying the port dimensions.  The trace
            dimensions are given by the rectangle perpendicular to the
            `excitation_axis` and at the max value for that excitation
            axis.  The excitation will dimensions are from the minimum
            `exctitation_axis` value to the maximum value.  The order
            of coordinates does not matter.
        :param propagation_axis: Specifies the coordinate axis and
            direction in which the signal propagation occurs.
        :param excitation_axis: Axis and direction of signal
            excitation.
        :param thickness: Metal trace thickness.  Units are whatever
            you set the CSX unit to, defaults to m.
        :param conductivity: Metal conductivity (in S/m).  The default
            uses the conductivity of copper.
        :param feed_impedance: The feeding impedance value.  The
            default value of None creates an infinite impedance.  If
            you use the default value ensure that the port is
            terminated by a PML.  When performing a characteristic
            impedance measurement use the default value and PML, which
            gives better results than attempting to use a matching
            impedance.
        :param feed_shift: The amount by which to shift the feed as a
            fraction of the total port length.  The final position
            will be influenced by this value but adjusted for the mesh
            used.
        :param ref_impedance: The impedance used to calculate the port
            voltage and current values.  If left as the default value
            of None, the calculated characteristic impedance is used
            to calculate these values.
        :param measurement_shift: The amount by which to shift the
            measurement probes as a fraction of the total port length.
            By default, the measurement port is placed halfway between
            the start and stop.  Like `feed_shift`, the final position
            will be adjusted for the mesh used.  This is important
            since voltage probes need to lie on mesh lines and current
            probes need to be placed equidistant between them.
        """
        super().__init__(sim=sim, number=number, excite=excite)
        self._box = box
        self._box.set_increasing()
        self._propagation_axis = propagation_axis
        self._excitation_axis = excitation_axis
        self._check_axes_perpendicular()
        self.thickness = thickness
        self.conductivity = conductivity
        self.feed_impedance = feed_impedance
        self.feed_shift = feed_shift
        self._ref_impedance = ref_impedance
        self.measurement_shift = measurement_shift

        self._set_trace()

    @property
    def box(self) -> Box3:
        """
        """
        return self._box

    def propagation_axis(self) -> Axis:
        """
        """
        return self._propagation_axis

    def get_feed_shift(self) -> float:
        """
        """
        return self.feed_shift

    def _check_axes_perpendicular(self) -> None:
        """
        """
        if self._propagation_axis.axis == self._excitation_axis.axis:
            raise ValueError(
                "Excitation and propagation axes must be perpendicular."
            )

    def _set_trace(self) -> None:
        """
        Set trace.
        """
        trace_prop = self._sim.csx.AddConductingSheet(
            "Microstrip_Trace_" + str(self.number),
            conductivity=self.conductivity,
            thickness=self.thickness,
        )
        box = self._trace_box()
        trace_prop.AddBox(
            priority=priorities["trace"], start=box.start(), stop=box.stop(),
        )

    def _trace_box(self) -> Box3:
        """
        Get the trace box.
        """
        trace_box = deepcopy(self.box)
        excitation_axis = self._excitation_axis.axis
        if self._excitation_axis.is_positive_direction():
            trace_box.min_corner[excitation_axis] = trace_box.max_corner[
                excitation_axis
            ]
        else:
            trace_box.max_corner[excitation_axis] = trace_box.min_corner[
                excitation_axis
            ]

        return trace_box

    def _propagation_direction(self) -> int:
        """
        Get the direction of the signal propagation.
        """
        return self._propagation_axis.direction

    def _excitation_direction(self) -> int:
        """
        Get the direction of the signal excitation.
        """
        return self._excitation_axis.direction

    def _trace_perpendicular_axis(self) -> Axis:
        """
        """
        axes = [0, 1, 2]
        axes.remove(self._propagation_axis.axis)
        axes.remove(self._excitation_axis.axis)
        trace_perp_axis = axes[0]
        if self._propagation_axis.is_positive_direction():
            return Axis(trace_perp_axis)
        else:
            return Axis(trace_perp_axis, -1)

    def _set_feed(self, mesh: Mesh) -> None:
        """
        Set excitation feed.
        """
        excite_type = None
        if self.excite:
            excite_type = 0

        feed = Feed(
            sim=self._sim,
            box=self._feed_box(mesh),
            excite_direction=self._excitation_axis.as_list(),
            excite_type=excite_type,
            impedance=self.feed_impedance,
        )
        self.feeds.append(feed)

    def _set_probes(self, mesh: Mesh) -> None:
        """
        Set measurement probes.
        """
        trace_box = deepcopy(self._trace_box())
        trace_perp_axis = self._trace_perpendicular_axis().axis
        trace_perp_low = trace_box.min_corner[trace_perp_axis]
        trace_perp_high = trace_box.max_corner[trace_perp_axis]
        trace_perp_mid = np.average([trace_perp_low, trace_perp_high])

        excitation_axis = self._excitation_axis.axis
        if self._excitation_axis.is_positive_direction():
            gnd_pos = self.box.min_corner[excitation_axis]
            trace_pos = self.box.max_corner[excitation_axis]
        else:
            gnd_pos = self.box.max_corner[excitation_axis]
            trace_pos = self.box.min_corner[excitation_axis]

        prop_axis = self._propagation_axis.axis
        trace_prop_low = trace_box.min_corner[prop_axis]
        trace_prop_high = trace_box.max_corner[prop_axis]
        if self._propagation_axis.is_positive_direction():
            prop_index, vxmid = mesh.nearest_mesh_line(
                prop_axis,
                trace_box.min_corner[prop_axis]
                + (
                    self.measurement_shift * (trace_prop_high - trace_prop_low)
                ),
            )
        else:
            prop_index, vxmid = mesh.nearest_mesh_line(
                prop_axis,
                trace_box.max_corner[prop_axis]
                - (
                    self.measurement_shift * (trace_prop_high - trace_prop_low)
                ),
            )
        mesh.set_lines_equidistant(0, prop_index - 1, prop_index + 1)

        v_prop_pos = [
            mesh.get_mesh_line(
                prop_axis, prop_index - self._propagation_direction()
            ),
            mesh.get_mesh_line(prop_axis, prop_index),
            mesh.get_mesh_line(
                prop_axis, prop_index + self._propagation_direction()
            ),
        ]
        i_prop_pos = [
            (v_prop_pos[0] + v_prop_pos[1]) / 2,
            (v_prop_pos[1] + v_prop_pos[2]) / 2,
        ]

        for idx in range(3):
            box = Box3(
                Coordinate3(None, None, None), Coordinate3(None, None, None)
            )
            box.min_corner[prop_axis] = v_prop_pos[idx]
            box.max_corner[prop_axis] = v_prop_pos[idx]
            box.min_corner[trace_perp_axis] = trace_perp_mid
            box.max_corner[trace_perp_axis] = trace_perp_mid
            box.min_corner[excitation_axis] = gnd_pos
            box.max_corner[excitation_axis] = trace_pos
            self.vprobes.append(
                Probe(
                    sim=self._sim,
                    box=box,
                    p_type=0,
                    weight=self._excitation_direction(),
                )
            )

        for idx in range(2):
            box = Box3(
                Coordinate3(None, None, None), Coordinate3(None, None, None)
            )
            box.min_corner[prop_axis] = i_prop_pos[idx]
            box.max_corner[prop_axis] = i_prop_pos[idx]
            box.min_corner[trace_perp_axis] = trace_perp_low
            box.max_corner[trace_perp_axis] = trace_perp_high
            box.min_corner[excitation_axis] = trace_pos
            box.max_corner[excitation_axis] = trace_pos
            self.iprobes.append(
                Probe(
                    sim=self._sim,
                    box=box,
                    p_type=1,
                    normal_axis=self._propagation_axis,
                    weight=-self._propagation_direction(),  # TODO negative??
                )
            )

    def _feed_box(self, mesh: Mesh) -> Box3:
        """
        Get the excitation feed box.
        """
        box = deepcopy(self.box)
        feed_axis = self._excitation_axis.axis
        if not self._excitation_axis.is_positive_direction():
            old_max = box.max_corner[feed_axis]
            box.max_corner[feed_axis] = box.min_corner[feed_axis]
            box.min_corner[feed_axis] = old_max

        prop_axis = self._propagation_axis.axis
        prop_dist = (
            self.box.max_corner[prop_axis] - self.box.min_corner[prop_axis]
        )
        if self._propagation_axis.is_positive_direction():
            _, prop_pos = mesh.nearest_mesh_line(
                prop_axis,
                self.box.min_corner[prop_axis] + (self.feed_shift * prop_dist),
            )
        else:
            _, prop_pos = mesh.nearest_mesh_line(
                prop_axis,
                self.box.max_corner[prop_axis] - (self.feed_shift * prop_dist),
            )

        box.max_corner[prop_axis] = prop_pos
        box.min_corner[prop_axis] = prop_pos

        return box


class DifferentialMicrostripPort(Port):
    """
    """

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        propagation_axis: Axis,
        excitation_axis: Axis,
        number: int,
        gap: float,
        thickness: float,
        conductivity: float = 5.8e7,
        excite: bool = False,
        feed_impedance: float = None,
        feed_shift: float = 0.2,
        ref_impedance: float = None,
        measurement_shift: float = 0.5,
    ):
        """
        :param sim: Simulation to which microstrip port is added.
        :param box: 3D box specifying the port dimensions.  This box
            fully encompasses both microstrip traces.  The length of
            each trace is identical and is equal to the length of the
            box in the direction of `propagation_axis`.  The width of
            the box in the direction of `excitation_axis` gives the
            distance between the outer edges of each trace.  The size
            of the box in the direction perpendicular to these two
            gives the elevation of the differential pair and must have
            size 0.  The order of coordinates in the box does not
            matter, since the direction of the port is set by the
            direction of `propagation_axis`.
        :param propagation_axis: Specifies the coordinate axis and
            direction in which the signal propagation occurs.
        :param excitation_axis: Axis and direction of signal
            excitation.  This should point from one trace to the other
            trace.  In other words, `propagation_axis` and
            `excitation_axis` (along with the box elevation) together
            determine the plane in which the differential traces
            reside.
        :param gap: Separation between inner edges of microstrip
            traces.
        :param thickness: Metal trace thickness.  Units are whatever
            you set the CSX unit to, defaults to m.
        :param conductivity: Metal conductivity (in S/m).  The default
            uses the conductivity of copper.
        :param feed_impedance: The feeding impedance value.  The
            default value of None creates an infinite impedance.  If
            you use the default value ensure that the port is
            terminated by a PML.  When performing a characteristic
            impedance measurement use the default value and PML, which
            gives better results than attempting to use a matching
            impedance.
        :param feed_shift: The amount by which to shift the feed as a
            fraction of the total port length.  The final position
            will be influenced by this value but adjusted for the mesh
            used.
        :param ref_impedance: The impedance used to calculate the port
            voltage and current values.  If left as the default value
            of None, the calculated characteristic impedance is used
            to calculate these values.
        :param measurement_shift: The amount by which to shift the
            measurement probes as a fraction of the total port length.
            By default, the measurement port is placed halfway between
            the start and stop.  Like `feed_shift`, the final position
            will be adjusted for the mesh used.  This is important
            since voltage probes need to lie on mesh lines and current
            probes need to be placed equidistant between them.
        """
        super().__init__(sim=sim, number=number, excite=excite)
        self._box = box
        self._box.set_increasing()
        self._propagation_axis = propagation_axis
        self._excitation_axis = excitation_axis
        self._check_axes_perpendicular()
        self._check_normal_axis_size()
        self._gap = gap
        self._thickness = thickness
        self._conductivity = conductivity
        self._feed_impedance = feed_impedance
        self._feed_shift = feed_shift
        self._ref_impedance = ref_impedance
        self._measurement_shift = measurement_shift

        self._set_traces()

    def propagation_axis(self) -> Axis:
        """
        """
        return self._propagation_axis

    def _set_traces(self) -> None:
        """
        """
        trace_prop = self._sim.csx.AddConductingSheet(
            "Differential_Microstrip_Trace_" + str(self.number),
            conductivity=self._conductivity,
            thickness=self._thickness,
        )
        for box in self._trace_boxes():
            trace_prop.AddBox(
                priority=priorities["trace"],
                start=box.start(),
                stop=box.stop(),
            )

    def _trace_boxes(self) -> Tuple[Box3, Box3]:
        """
        """
        lower_box = deepcopy(self._box)
        upper_box = deepcopy(self._box)
        excite_axis = self._excitation_axis.axis
        trace_width = self._trace_width()

        lower_box.max_corner[excite_axis] = (
            lower_box.min_corner[excite_axis] + trace_width
        )
        upper_box.min_corner[excite_axis] = (
            upper_box.max_corner[excite_axis] - trace_width
        )

        return (lower_box, upper_box)

    def _trace_width(self) -> float:
        """
        """
        box_width = (
            self._box.max_corner[self._excitation_axis.axis]
            - self._box.min_corner[self._excitation_axis.axis]
        )
        return (box_width - self._gap) / 2

    def _normal_axis(self) -> Axis:
        """
        """
        axes = [0, 1, 2]
        axes.remove(self._propagation_axis.axis)
        axes.remove(self._excitation_axis.axis)
        return Axis(axes[0])

    def _check_axes_perpendicular(self) -> None:
        """
        """
        if self._propagation_axis.axis == self._excitation_axis.axis:
            raise ValueError(
                "Excitation and propagation axes must be perpendicular."
            )

    def _check_normal_axis_size(self) -> None:
        """
        """
        normal_axis = self._normal_axis().axis
        if (
            self._box.max_corner[normal_axis]
            != self._box.min_corner[normal_axis]
        ):
            raise ValueError(
                "Size of box in direction normal to microstrip plane must be 0."
            )

    def _propagation_direction(self) -> int:
        """
        Get the direction of the signal propagation.
        """
        return self._propagation_axis.direction

    def _excitation_direction(self) -> int:
        """
        Get the direction of the signal excitation.
        """
        return self._excitation_axis.direction

    def _set_feed(self, mesh: Mesh) -> None:
        """
        Set excitation feed.
        """
        excite_type = None
        if self.excite:
            excite_type = 0

        feed = Feed(
            sim=self._sim,
            box=self._feed_box(mesh),
            excite_direction=self._excitation_axis.as_list(),
            excite_type=excite_type,
            impedance=self._feed_impedance,
        )
        self.feeds.append(feed)

    def _feed_box(self, mesh: Mesh) -> Box3:
        """
        Get the excitation feed box.
        """
        box = deepcopy(self._box)
        feed_axis = self._excitation_axis.axis
        trace_width = self._trace_width()
        box.max_corner[feed_axis] -= trace_width
        box.min_corner[feed_axis] += trace_width
        if not self._excitation_axis.is_positive_direction():
            old_max = box.max_corner[feed_axis]
            box.max_corner[feed_axis] = box.min_corner[feed_axis]
            box.min_corner[feed_axis] = old_max

        prop_axis = self._propagation_axis.axis
        prop_dist = (
            self._box.max_corner[prop_axis] - self._box.min_corner[prop_axis]
        )
        if self._propagation_axis.is_positive_direction():
            _, prop_pos = mesh.nearest_mesh_line(
                prop_axis,
                self._box.min_corner[prop_axis]
                + (self._feed_shift * prop_dist),
            )
        else:
            _, prop_pos = mesh.nearest_mesh_line(
                prop_axis,
                self._box.max_corner[prop_axis]
                - (self._feed_shift * prop_dist),
            )

        box.max_corner[prop_axis] = prop_pos
        box.min_corner[prop_axis] = prop_pos

        return box

    def _set_probes(self, mesh: Mesh) -> None:
        """
        Set measurement probes.
        """
        prop_axis = self._propagation_axis.axis
        excite_axis = self._excitation_axis.axis
        normal_axis = self._normal_axis().axis
        normal_pos = self._box.min_corner[normal_axis]
        trace_width = self._trace_width()
        excite_pos_lower = self._box.min_corner[excite_axis] + trace_width
        excite_pos_upper = self._box.max_corner[excite_axis] - trace_width

        prop_dist = (
            self._box.max_corner[prop_axis] - self._box.min_corner[prop_axis]
        )

        if self._propagation_axis.is_positive_direction():
            prop_index, _ = mesh.nearest_mesh_line(
                prop_axis,
                self._box.min_corner[prop_axis]
                + (self._measurement_shift * prop_dist),
            )
        else:
            prop_index, _ = mesh.nearest_mesh_line(
                prop_axis,
                self._box.max_corner[prop_axis]
                - (self._measurement_shift * prop_dist),
            )
        mesh.set_lines_equidistant(0, prop_index - 1, prop_index + 1)

        v_prop_pos = [
            mesh.get_mesh_line(
                prop_axis, prop_index - self._propagation_direction()
            ),
            mesh.get_mesh_line(prop_axis, prop_index),
            mesh.get_mesh_line(
                prop_axis, prop_index + self._propagation_direction()
            ),
        ]
        i_prop_pos = [
            (v_prop_pos[0] + v_prop_pos[1]) / 2,
            (v_prop_pos[1] + v_prop_pos[2]) / 2,
        ]

        for idx in range(3):
            box = Box3(
                Coordinate3(None, None, None), Coordinate3(None, None, None)
            )
            box.min_corner[prop_axis] = v_prop_pos[idx]
            box.max_corner[prop_axis] = v_prop_pos[idx]
            box.min_corner[normal_axis] = normal_pos
            box.max_corner[normal_axis] = normal_pos
            box.min_corner[excite_axis] = excite_pos_lower
            box.max_corner[excite_axis] = excite_pos_upper
            self.vprobes.append(
                Probe(
                    sim=self._sim,
                    box=box,
                    p_type=0,
                    weight=self._excitation_direction(),
                )
            )

        for idx in range(2):
            boxes = [None, None]
            boxes[0] = Box3(
                Coordinate3(None, None, None), Coordinate3(None, None, None)
            )
            boxes[1] = deepcopy(boxes[0])

            for box in boxes:
                box.min_corner[prop_axis] = i_prop_pos[idx]
                box.max_corner[prop_axis] = i_prop_pos[idx]
                box.min_corner[normal_axis] = normal_pos
                box.max_corner[normal_axis] = normal_pos

            boxes[0].min_corner[excite_axis] = (
                self._box.max_corner[excite_axis] - trace_width
            )
            boxes[0].max_corner[excite_axis] = self._box.max_corner[
                excite_axis
            ]
            boxes[1].min_corner[excite_axis] = self._box.min_corner[
                excite_axis
            ]
            boxes[1].max_corner[excite_axis] = (
                self._box.min_corner[excite_axis] + trace_width
            )

            for i, box in enumerate(boxes):
                self.iprobes.append(
                    Probe(
                        sim=self._sim,
                        box=box,
                        p_type=1,
                        normal_axis=self._propagation_axis,
                        weight=-self._propagation_direction()
                        * ((i + 1) ** -1),  # TODO negative prop direction??
                    )
                )

    def _probe_vi(self) -> Tuple[np.array, np.array, np.array, np.array]:
        """
        Read probe voltage and current values.

        :returns: (v, i, dv, di)
        """
        [vprobe.read() for vprobe in self.vprobes]
        [iprobe.read() for iprobe in self.iprobes]
        prop_axis = self.propagation_axis().axis
        self._data_read = True
        v = self.vprobes[1].get_freq_data()[1]
        # TODO using both current probes gives slightly incorrect
        # results for some unknown reason.
        # # the currents on each trace should be exactly opposite, but
        # # we use both just in case.
        # i1 = 0.5 * (
        #     self.iprobes[0].get_freq_data()[1]
        #     - self.iprobes[1].get_freq_data()[1]
        # )
        # i2 = 0.5 * (
        #     self.iprobes[2].get_freq_data()[1]
        #     - self.iprobes[3].get_freq_data()[1]
        # )
        # i = 0.5 * (i1 + i2)
        i = 0.5 * (
            self.iprobes[0].get_freq_data()[1]
            + self.iprobes[2].get_freq_data()[1]
        )
        dv = (
            self.vprobes[2].get_freq_data()[1]
            - self.vprobes[0].get_freq_data()[1]
        ) / (
            self.sim.unit
            * np.abs(
                self.vprobes[2].box.min_corner[prop_axis]
                - self.vprobes[0].box.min_corner[prop_axis]
            )
        )
        # di = (i2 - i1) / (
        #     self.sim.unit
        #     * np.abs(
        #         self.iprobes[2].box.min_corner[prop_axis]
        #         - self.iprobes[0].box.min_corner[prop_axis]
        #     )
        # )
        di = (
            self.iprobes[2].get_freq_data()[1]
            - self.iprobes[0].get_freq_data()[1]
        ) / (
            self.sim.unit
            * np.abs(
                self.iprobes[2].box.min_corner[prop_axis]
                - self.iprobes[0].box.min_corner[prop_axis]
            )
        )

        return (v, i, dv, di)


class CPWPort(Port):
    """
    Coplanar waveguide transmission line port.
    """

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        gap: float,
        thickness: float,
        conductivity: float = 5.8e7,
        excite: bool = False,
        feed_impedance: float = None,
        feed_shift: float = 0.2,
        ref_impedance: float = None,
        measurement_shift: float = 0.5,
    ):
        """
        :param gap: Gap between adjacent ground planes and trace (in m).
        """
        self.gap = gap
        raise RuntimeError(
            "TODO CPWPort is not correctly implemented. "
            "See MicrostripPort for guidance on how to implement correctly."
        )
        super().__init__(
            sim=sim,
            box=box,
            thickness=thickness,
            conductivity=conductivity,
            excite=excite,
            feed_impedance=feed_impedance,
            feed_shift=feed_shift,
            ref_impedance=ref_impedance,
            measurement_shift=measurement_shift,
        )

    def calc(self) -> None:
        """
        Calculate the characteristic impedance, propagation constant,
        and incident and reflected power.
        """
        [vprobe.read() for vprobe in self.vprobes]
        [iprobe.read() for iprobe in self.iprobes]
        self._data_read = True
        v0 = 0.5 * (
            self.vprobes[0].get_freq_data()[1]
            + self.vprobes[1].get_freq_data()[1]
        )
        v1 = 0.5 * (
            self.vprobes[2].get_freq_data()[1]
            + self.vprobes[3].get_freq_data()[1]
        )
        v2 = 0.5 * (
            self.vprobes[4].get_freq_data()[1]
            + self.vprobes[5].get_freq_data()[1]
        )
        v = v1

        i = 0.5 * (
            self.iprobes[0].get_freq_data()[1]
            + self.iprobes[1].get_freq_data()[1]
        )
        dv = (v2 - v0) / (
            self.sim.unit
            * (
                self.vprobes[2].box.min_corner.x
                - self.vprobes[0].box.min_corner.x
            )
        )
        di = (
            self.iprobes[1].get_freq_data()[1]
            - self.iprobes[0].get_freq_data()[1]
        ) / (
            self.sim.unit
            * (
                self.iprobes[1].box.min_corner.x
                - self.iprobes[0].box.min_corner.x
            )
        )

        self._calc_beta(v, i, dv, di)
        self._calc_z0(v, i, dv, di)
        self._calc_v_inc(v, i)
        self._calc_v_ref(v, i)
        self._calc_i_inc(v, i)
        self._calc_i_ref(v, i)
        self._calc_p_inc()
        self._calc_p_ref()

    def _set_feed(self, mesh: Mesh) -> None:
        """
        Set excitation feed.
        """
        excite_type = None
        if self.excite:
            excite_type = 0

        # when the user specifies a feed impedance, this means they
        # want the cpw port terminated with that impedance. To get
        # the best termination, use a single feed extending the trace
        # line. The user must provide a ground plane 1 gap width
        # behind the trace end.
        if self.feed_impedance:
            trace_box = self._trace_box()
            feed = Feed(
                sim=self._sim,
                box=Box3(
                    Coordinate3(
                        trace_box.min_corner.x
                        - (self._propagation_direction() * self.gap),
                        trace_box.min_corner.y,
                        trace_box.min_corner.z,
                    ),
                    Coordinate3(
                        trace_box.min_corner.x,
                        trace_box.max_corner.y,
                        trace_box.max_corner.z,
                    ),
                ),
                excite_direction=[self._propagation_direction(), 0, 0],
                excite_type=excite_type,
                impedance=self.feed_impedance,
            )
            self.feeds.append(feed)

        else:
            for box, excite_dir in zip(
                self._get_feed_boxes(mesh), [[0, 1, 0], [0, -1, 0]]
            ):
                feed = Feed(
                    sim=self._sim,
                    box=box,
                    excite_direction=excite_dir,
                    excite_type=excite_type,
                    impedance=self.feed_impedance,
                )
                self.feeds.append(feed)

    def _get_feed_boxes(self, mesh: Mesh) -> None:
        """
        """
        _, xpos = mesh.nearest_mesh_line(
            0,
            self.box.min_corner.x
            + (
                self.feed_shift
                * (self.box.max_corner.x - self.box.min_corner.x)
            ),
        )
        feed_boxes = [
            Box3(Coordinate3(xpos, ystart, 0), Coordinate3(xpos, yend, 0))
            for ystart, yend in zip(
                [
                    self.box.min_corner.y
                    - (self._propagation_direction() * self.gap),
                    self.box.max_corner.y
                    + (self._propagation_direction() * self.gap),
                ],
                [self.box.min_corner.y, self.box.max_corner.y],
            )
        ]
        return feed_boxes

    def _set_probes(self, mesh: Mesh) -> None:
        """
        Set measurement probes.
        """
        trace_box = self._trace_box()
        trace_ylow = trace_box.min_corner.y
        trace_yhigh = trace_box.max_corner.y
        trace_z = trace_box.max_corner.z

        trace_xlow = trace_box.min_corner.x
        trace_xhigh = trace_box.max_corner.x
        x_index, vxmid = mesh.nearest_mesh_line(
            0,
            trace_box.min_corner.x
            + (self.measurement_shift * (trace_xhigh - trace_xlow)),
        )
        mesh.set_lines_equidistant(0, x_index - 1, x_index + 1)

        vxpos = [
            mesh.get_mesh_line(0, x_index - self._propagation_direction()),
            mesh.get_mesh_line(0, x_index),
            mesh.get_mesh_line(0, x_index + self._propagation_direction()),
        ]
        ixpos = [
            (vxpos[0] + vxpos[1]) / 2,
            (vxpos[1] + vxpos[2]) / 2,
        ]

        self.vprobes = []
        for xpos in vxpos:
            # TODO ensure probe weights set correctly
            self.vprobes.append(
                Probe(
                    sim=self._sim,
                    box=Box3(
                        Coordinate3(xpos, trace_ylow - self.gap, trace_z),
                        Coordinate3(xpos, trace_ylow, trace_z),
                    ),
                    p_type=0,
                    weight=self._excitation_direction(),
                )
            )
            self.vprobes.append(
                Probe(
                    sim=self._sim,
                    box=Box3(
                        Coordinate3(xpos, trace_yhigh, trace_z),
                        Coordinate3(xpos, trace_yhigh + self.gap, trace_z),
                    ),
                    p_type=0,
                    weight=-self._excitation_direction(),
                )
            )

        self.iprobes = [
            Probe(
                sim=self._sim,
                box=Box3(
                    Coordinate3(xpos, trace_ylow, trace_z),
                    Coordinate3(xpos, trace_yhigh, trace_z),
                ),
                p_type=1,
                normal_axis=Axis("x"),
                weight=-self._propagation_direction(),
            )
            for xpos in ixpos
        ]


class RectWaveguidePort(Port):
    """
    Rectangular waveguide port base class.
    """

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        propagation_axis: Axis,
        mode_name: str = "TE10",
        excite: bool = False,
        delay: float = 0,
        ref_impedance: float = None,
    ):
        """
        :param box: Waveguide box.  The waveguide width and height are
            computed from this box.
        :param propagation_axis: Direction the waveguide is facing.
            Set to 0, 1, or 2, for x, y, or z.
        :param mode_name: Waveguide propagation mode.  If you don't
            know this you probably want the default.  Otherwise, see
            Pozar (4e) p.110 for more information.
        :param excite: If True, add an excitation to the port.
        """
        super().__init__(sim, excite)
        self._box = box
        self._propagation_axis = propagation_axis
        self.mode_name = mode_name
        self.excite = excite
        self.delay = delay
        self._ref_impedance = ref_impedance

        # set later
        self.te = None
        self.e_func = [0, 0, 0]
        self.h_func = [0, 0, 0]
        self.kc = None
        self.direction = None
        self.a = None
        self.b = None

        self._set_width_height()
        self._set_direction()
        self._parse_mode_name()
        self._set_func()

    def propagation_axis(self) -> Axis:
        """
        """
        return self._propagation_axis

    @property
    def box(self) -> Box3:
        """
        """
        return self._box

    def calc(self):
        """
        Calculate the characteristic impedance, propagation constant,
        and incident and reflected voltage.

        TODO this doesn't account for reference plane shifts.  See
        p.184 of Pozar, or original openems ports.py file for this.
        """
        k = wavenumber(self.sim.freq, 1)  # use meters
        self._calc_cutoff_wavenumber()
        self._calc_beta(k)
        self._calc_z0(k)
        [vprobe.read() for vprobe in self.vprobes]
        [iprobe.read() for iprobe in self.iprobes]
        self._data_read = True
        v = self.vprobes[0].get_freq_data()[1]
        i = self.iprobes[0].get_freq_data()[1]
        self._calc_v_inc(v, i)
        self._calc_v_ref(v, i)
        self._calc_i_inc(v, i)
        self._calc_i_ref(v, i)
        self._calc_p_inc()
        self._calc_p_ref()

    def add_metal_shell(self, thickness: float) -> None:
        """
        """
        shell_prop = self._sim.csx.AddMetal("rect_wg_metal")
        back_face = self._shell_face_box(
            const_dim=self.propagation_axis().intval(),
            const_dim_idx=0,
            thickness=thickness,
        )
        shell_prop.AddBox(start=back_face.start(), stop=back_face.stop())
        dims = list(range(3))
        del dims[self.propagation_axis().intval()]
        for dim in dims:
            for i in range(2):
                face = self._shell_face_box(
                    const_dim=dim, const_dim_idx=i, thickness=thickness
                )
                shell_prop.AddBox(start=face.start(), stop=face.stop())

    def _shell_face_box(
        self, const_dim: int, const_dim_idx: int, thickness: float,
    ) -> Box3:
        """
        """
        box = Box3(
            Coordinate3(None, None, None), Coordinate3(None, None, None)
        )
        for dim in range(3):
            if dim == const_dim:
                box.min_corner[dim] = self.box[const_dim_idx][dim] + (
                    self._dim_outer_dir(dim, const_dim_idx) * thickness
                )
                box.max_corner[dim] = self.box[const_dim_idx][dim]
            else:
                box.min_corner[dim] = self.box.min_corner[dim] - thickness
                box.max_corner[dim] = self.box.max_corner[dim] + thickness

        return box

    def _dim_outer_dir(self, dim: int, dim_idx: int) -> int:
        """
        """
        other_dim_idx = (dim_idx + 1) % 2
        return np.sign(self.box[dim_idx][dim] - self.box[other_dim_idx][dim])

    def _set_width_height(self) -> None:
        """
        Set the waveguide width (a) and height (b) based on the box
        dimensions and propagation axis.  The width is always taken to
        be the larger of the two.
        """
        dimensions = [
            np.abs(self.box.max_corner[dim] - self.box.min_corner[dim])
            for dim in range(3)
        ]
        del dimensions[self.propagation_axis().intval()]
        self.a = np.amax(dimensions)
        self.b = np.amin(dimensions)

    def _parse_mode_name(self) -> None:
        """
        Parse mode_name argument to extract mode information.
        """
        assert len(self.mode_name) == 4, "Invalid mode definition"
        if self.mode_name.startswith("TE"):
            self.te = True
        else:
            self.te = False
        self.m = float(self.mode_name[2])
        self.n = float(self.mode_name[3])

        assert (
            self.te
        ), "Currently only TE-modes are supported! Mode found: {}".format(
            self.mode_name
        )

    def _set_direction(self) -> None:
        """
        Compute whether the port faces in the positive or negative
        direction of propagation_axis.
        """
        self.direction = int(
            np.sign(
                self.box.max_corner[self.propagation_axis().intval()]
                - self.box.min_corner[self.propagation_axis().intval()]
            )
        )

    def _set_func(self) -> None:
        """
        """
        ny_p = (self.propagation_axis().intval() + 1) % 3
        ny_pp = (self.propagation_axis().intval() + 2) % 3

        xyz = "xyz"
        if self.box.min_corner[ny_p] != 0:
            name_p = "({}-{})".format(xyz[ny_p], self.box.min_corner[ny_p])
        else:
            name_p = xyz[ny_p]
        if self.box.min_corner[ny_pp] != 0:
            name_pp = "({}-{})".format(xyz[ny_p], self.box.min_corner[ny_p])
        else:
            name_pp = xyz[ny_p]

        if self.n > 0:
            self.e_func[ny_p] = "{}*cos({}*{})*sin({}*{})".format(
                self.n / self.b,
                self.m * np.pi / self.a,
                name_p,
                self.n * np.pi / self.b,
                name_pp,
            )
        if self.m > 0:
            self.e_func[ny_pp] = "{}*sin({}*{})*cos({}*{})".format(
                -self.m / self.a,
                self.m * np.pi / self.a,
                name_p,
                self.n * np.pi / self.b,
                name_pp,
            )

        if self.m > 0:
            self.h_func[ny_p] = "{}*sin({}*{})*cos({}*{})".format(
                self.m / self.a,
                self.m * np.pi / self.a,
                name_p,
                self.n * np.pi / self.b,
                name_pp,
            )
        if self.n > 0:
            self.h_func[ny_pp] = "{}*cos({}*{})*sin({}*{})".format(
                self.n / self.b,
                self.m * np.pi / self.a,
                name_p,
                self.n * np.pi / self.b,
                name_pp,
            )

    def _calc_beta(self, k) -> None:
        """
        Calculate the propagation constant.
        """
        self.beta = np.sqrt(np.power(k, 2) - np.power(self.kc, 2))

    def _calc_z0(self, k) -> None:
        """
        Calculate the characteristic impedance.
        """
        self.z0 = k * Z0 / self.beta
        if self._ref_impedance is None:
            self._ref_impedance = self.z0

    def _set_probes(self, mesh: Mesh) -> None:
        """
        Set measurement probes.
        """
        _, prop_pos = mesh.nearest_mesh_line(
            self.propagation_axis().intval(),
            self.box[1][self.propagation_axis().intval()],
        )

        probe_box = deepcopy(self.box)
        probe_box.min_corner[self.propagation_axis().intval()] = prop_pos
        probe_box.max_corner[self.propagation_axis().intval()] = prop_pos

        self.vprobes = [
            Probe(
                sim=self._sim,
                box=probe_box,
                p_type=10,
                mode_function=self.e_func,
            )
        ]

        self.iprobes = [
            Probe(
                sim=self._sim,
                box=probe_box,
                p_type=11,
                weight=self.direction,
                mode_function=self.h_func,
            )
        ]

    def _set_feed(self, mesh: Mesh) -> None:
        """
        Set excitation feed.
        """
        if self.excite:
            _, prop_pos = mesh.nearest_mesh_line(
                self.propagation_axis().intval(),
                self.box[0][self.propagation_axis().intval()],
            )

            feed_box = deepcopy(self.box)
            feed_box.min_corner[self.propagation_axis().intval()] = prop_pos
            feed_box.max_corner[self.propagation_axis().intval()] = prop_pos

            feed_vec = np.ones(3)
            feed_vec[self.propagation_axis().intval()] = 0
            weight_func = [str(x) for x in self.e_func]
            feed = Feed(
                sim=self._sim,
                box=feed_box,
                excite_direction=feed_vec,
                excite_type=0,
                weight_func=weight_func,
                delay=self.delay,
            )
            self.feeds.append(feed)

    def _calc_cutoff_wavenumber(self) -> None:
        """
        Calculate the minimum wavenumber for wave to propagate.

        See Pozar (4e) p.112 for derivation.
        """
        a = self.a * self.sim.unit
        b = self.b * self.sim.unit
        self.kc = np.sqrt(
            np.power((self.m * np.pi / a), 2)
            + np.power((self.n * np.pi / b), 2)
        )


class CoaxPort(Port):
    """
    Coaxial port.  This does not construct a coaxial cable.  For that,
    use Coax in structure.py.
    """

    def __init__(
        self,
        sim: Simulation,
        number: int,
        start: Coordinate3,
        stop: Coordinate3,
        radius: float,
        core_radius: float,
        excite: bool = False,
        feed_shift: float = 0.2,
        feed_impedance: float = None,
        measurement_shift: float = 0.5,
        delay: float = 0,
        ref_impedance: float = None,
    ):
        """
        :param start: Starting position of the coaxial port at its
            radial center.
        :param stop: Ending position of the coaxial port at its radial
            center.  `start` and `stop` can only differ in one
            dimension.  In other words, the coaxial port must be
            parallel to a coordinate axis.
        :param radius: Distance between center and outer conductor for
            a cross section of the coaxial cable.
        :param core_radius: Radius of the inner copper core.
        :param excite: If True, add an excitation to the port.
        :param feed_shift: Offsets feed from starting position.  Given
            as a proportion of the total length.
        :param feed_impedance: The feeding impedance value.  The
            default value of None creates an infinite impedance.  If
            you use the default value ensure that the port is
            terminated by a PML.  When performing a characteristic
            impedance measurement use the default value and PML, which
            gives better results than attempting to use a matching
            impedance.
        :param measurement_shift: Measurement probes offset.  Given as
            a proportion of the total length.
        :param delay:
        :param ref_impedance: Impedance used to calculate port
            parameters.  This does not affect the calculated
            characteristic impedance.
        """
        super().__init__(sim=sim, number=number, excite=excite)
        self._start = start
        self._stop = stop
        self._check_start_stop()
        self._radius = radius
        self._core_radius = core_radius
        self._excite = excite
        self._feed_shift = feed_shift
        self._feed_impedance = feed_impedance
        self._measurement_shift = measurement_shift
        self._delay = delay
        self._ref_impedance = ref_impedance

        self._set_core()

    def _check_start_stop(self) -> None:
        """
        """
        diff = np.diff(
            [self._stop.coordinate_list(), self._start.coordinate_list()],
            axis=0,
        )
        if np.count_nonzero(diff) != 1:
            raise ValueError(
                "Invalid start and stop coordinates. Port must be parallel "
                "to a coordinate axis."
            )

    def _propagation_axis(self) -> Axis:
        """
        """
        for i in range(3):
            if np.isclose(self._start[i], self._stop[i]):
                return Axis(i)
        raise RuntimeError("Unable to determine propagation axis.")

    def _direction(self) -> int:
        """
        Compute whether the port faces in the positive or negative
        direction of propagation_axis.
        """
        prop_axis = self._propagation_axis().axis
        direction = int(
            np.sign(self._stop[prop_axis] - self._start[prop_axis])
        )
        return direction

    def _set_core(self) -> None:
        """
        """
        # TODO bug?? cylinder ignored if start > stop
        if self._direction() == 1:
            start = self._start.coordinate_list()
            stop = self._stop.coordinate_list()
        else:
            start = self._stop.coordinate_list()
            stop = self._start.coordinate_list()

        core_prop = self.sim.csx.AddMetal(self._core_name())
        core_prop.AddCylinder(
            start=start,
            stop=stop,
            radius=self._core_radius,
            priority=priorities["trace"],
        )

    def propagation_axis(self) -> Axis:
        """
        """
        diff = np.diff(
            [self._stop.coordinate_list(), self._start.coordinate_list()],
            axis=0,
        )[0]
        for i in range(3):
            if diff[i] != 0:
                return Axis(i)

    def _direction(self) -> int:
        """
        +1 for positive direction in propagation axis direction and -1
        for negative direction.
        """
        axis_int = self.propagation_axis().intval()
        return int(np.sign(self._stop[axis_int] - self._start[axis_int]))

    def _core_name(self) -> str:
        """
        """
        return "Coax_core_" + str(self.number)

    def _set_probes(self, mesh: Mesh) -> None:
        """
        Set measurement probes.
        """
        prop_axis = self.propagation_axis().intval()
        pos = (
            self._measurement_shift
            * (self._stop[prop_axis] - self._start[prop_axis])
        ) + self._start[prop_axis]
        mid_idx, mid_pos = mesh.nearest_mesh_line(prop_axis, pos)
        mesh.set_lines_equidistant(prop_axis, mid_idx - 1, mid_idx + 1)
        low_idx = mid_idx - self._direction()
        high_idx = mid_idx + self._direction()

        vlow = mesh.get_mesh_line(prop_axis, low_idx)
        vmid = mesh.get_mesh_line(prop_axis, mid_idx)
        vhigh = mesh.get_mesh_line(prop_axis, high_idx)
        vpos = [vlow, vmid, vhigh]

        ilow = np.average([vlow, vmid])
        ihigh = np.average([vmid, vhigh])
        ipos = [ilow, ihigh]

        for pos in vpos:
            box = Box3(
                Coordinate3(None, None, None), Coordinate3(None, None, None)
            )
            box.min_corner[prop_axis] = pos
            box.max_corner[prop_axis] = pos

            other_axis1 = (prop_axis + 1) % 3
            box.min_corner[other_axis1] = (
                self._start[other_axis1] - self._radius
            )
            box.max_corner[other_axis1] = self._start[other_axis1]

            other_axis2 = (prop_axis + 2) % 3
            box.min_corner[other_axis2] = self._start[other_axis2]
            box.max_corner[other_axis2] = self._start[other_axis2]
            self.vprobes.append(
                Probe(sim=self.sim, box=box, p_type=0, weight=1)
            )

        for pos in ipos:
            box = Box3(
                Coordinate3(None, None, None), Coordinate3(None, None, None)
            )
            box.min_corner[prop_axis] = pos
            box.max_corner[prop_axis] = pos

            other_axis1 = (prop_axis + 1) % 3
            box.min_corner[other_axis1] = (
                self._start[other_axis1] - self._core_radius
            )
            box.max_corner[other_axis1] = (
                self._start[other_axis1] + self._core_radius
            )

            other_axis2 = (prop_axis + 2) % 3
            box.min_corner[other_axis2] = (
                self._start[other_axis2] - self._core_radius
            )
            box.max_corner[other_axis2] = (
                self._start[other_axis2] + self._core_radius
            )
            self.iprobes.append(
                Probe(
                    sim=self.sim,
                    box=box,
                    p_type=1,
                    normal_axis=self.propagation_axis(),
                    weight=-self._direction(),  # TODO negative??
                )
            )

    def _set_feed(self, mesh: Mesh) -> None:
        """
        Set excitation feed.
        """
        if self._excite:
            excite_type = 0
        else:
            excite_type = None

        prop_axis = self.propagation_axis().intval()
        pos = (
            self._direction()
            * self._feed_shift
            * (self._stop[prop_axis] - self._start[prop_axis])
        ) + self._start[prop_axis]
        _, feed_pos = mesh.nearest_mesh_line(prop_axis, pos)

        box = Box3(
            Coordinate3(None, None, None), Coordinate3(None, None, None)
        )
        box.min_corner[prop_axis] = feed_pos
        box.max_corner[prop_axis] = feed_pos

        other_axis1 = (prop_axis + 1) % 3
        box.min_corner[other_axis1] = self._start[other_axis1] - self._radius
        box.max_corner[other_axis1] = self._start[other_axis1]

        other_axis2 = (prop_axis + 2) % 3
        box.min_corner[other_axis2] = self._start[other_axis2]
        box.max_corner[other_axis2] = self._start[other_axis2]

        excite_direction = [0, 0, 0]
        excite_direction[other_axis1] = 1
        feed = Feed(
            sim=self.sim,
            box=box,
            excite_direction=excite_direction,
            excite_type=excite_type,
            impedance=self._feed_impedance,
            delay=self._delay,
        )
        self.feeds.append(feed)



from typing import List, Tuple
import numpy as np
from pyems.simulation import Simulation
from pyems.mesh import Mesh
from pyems.utilities import max_priority
from pyems.coordinate import Box3, box_overlap


class Feed:
    """
    Excitation feed.
    """

    unique_ctr = 0

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        excite_direction: List[float],
        excite_type: int = None,
        impedance: complex = None,
        weight_func=None,
        delay: int = 0,
    ):
        """
        :param sim: Simulation to which this feed is added.
        :param box: Feed box.
        :param excite_direction: The direction that the excitation
            propagates.  Provide a list of 3 values corresponding to
            x, y, and z.  For instance, [0, 0, 1] would propagate in
            the +z direction.
        :param excite_type: Excitation type.  See `SetExcitation`.
            Leave as the default None, if you don't want an
            excitation.
        :param impedance: Feed impedance.  If left as None, which is
            the default, the feed will have infinite impedance.  In
            this case make sure to terminate the structure in PMLs.
        :param weight_func: Excitation weighting function.  See
            `SetWeightFunction`.
        :param delay: Excitation delay in seconds.
        """
        self._sim = sim
        self._box = box
        self.impedance = impedance
        self.excite_direction = excite_direction
        self.excite_type = excite_type
        self.weight_func = weight_func
        self.delay = delay
        self.excitation_box = None
        self.res_box = None

        self.set_feed()

    @property
    def sim(self) -> Simulation:
        """
        """
        return self._sim

    @property
    def box(self) -> Box3:
        """
        """
        return self._box

    def set_feed(self) -> None:
        """
        Set excitation feed.
        """
        if self.excite_type is not None:
            excitation = self.sim.csx.AddExcitation(
                name="excite_" + str(self._get_inc_ctr()),
                exc_type=self.excite_type,
                exc_val=self.excite_direction,
                delay=self.delay,
            )
            if self.weight_func:
                excitation.SetWeightFunction(self.weight_func)

            self.excitation_box = excitation.AddBox(
                start=self.box.start(),
                stop=self.box.stop(),
                priority=max_priority(),
            )

        if self.impedance:
            rval, cval, lval = self._impedance_rcl()
            res = self.sim.csx.AddLumpedElement(
                name="resist_" + str(self._get_ctr()),
                ny=self._resist_dir(),
                caps=True,
                R=rval,
                C=cval,
                L=lval,
            )
            self.res_box = res.AddBox(
                start=self.box.start(), stop=self.box.stop()
            )

    def pml_overlap(self) -> bool:
        """
        """
        if self.excitation_box is None and self.res_box is None:
            return False

        pml_boxes = self.sim.mesh.pml_boxes()
        for pml_box in pml_boxes:
            if box_overlap(self.box, pml_box):
                return True
        return False

    def _impedance_rcl(self) -> Tuple[float, float, float]:
        """
        """
        if np.is_complex(self.impedance):
            raise RuntimeWarning(
                "Only feed resistances are currently supported."
            )

        return (np.real(self.impedance), 0, 0)

    def snap_to_mesh(self, mesh) -> None:
        """
        Align feed with the provided mesh.  It is necessary to call
        this function in order to get correct simulation results.

        :param mesh: Mesh object.
        """
        for dim in [0, 1, 2]:
            self._snap_dim(mesh, dim)

    def _snap_dim(self, mesh: Mesh, dim: int) -> None:
        """
        Align feed to mesh for a given dimension.  This function will
        only have an effect when the provided dimension has zero size.

        :param mesh: Mesh object.
        :param dim: Dimension.  0, 1, 2 for x, y, z.
        """
        if self.box.min_corner[dim] == self.box.max_corner[dim]:
            start = self.excitation_box.GetStart()
            stop = self.excitation_box.GetStop()
            _, pos = mesh.nearest_mesh_line(dim, start[dim])
            start[dim] = pos
            stop[dim] = pos
            self.excitation_box.SetStart(start)
            self.excitation_box.SetStop(stop)
            if self.impedance:
                self.res_box.SetStart(start)
                self.res_box.SetStop(stop)

    def _resist_dir(self) -> int:
        """
        AddLumpedElement requires a direction in the form of 0, 1, or
        2. Get this value from the excitation direction.
        """
        # TODO doesn't work when excite_direction has multiple directions.
        return abs(self.excite_direction[1] + (2 * self.excite_direction[2]))

    @classmethod
    def _get_ctr(cls):
        """
        """
        return cls.unique_ctr

    @classmethod
    def _inc_ctr(cls):
        """
        """
        cls.unique_ctr += 1

    @classmethod
    def _get_inc_ctr(cls):
        """
        """
        ctr = cls._get_ctr()
        cls._inc_ctr()
        return ctr



import sys
from bisect import bisect_left
from typing import List
import numpy as np
from CSXCAD.CSXCAD import ContinuousStructure
from CSXCAD.CSPrimitives import CSPrimitives
from CSXCAD.CSTransform import CSTransform


# TODO should set max precision instead of precision list. Precision
# should be computed automatically based on the last digit that
# differs between each value.
def print_table(
    data: np.array, col_names: List[str], prec: List[int], out_file=sys.stdout,
) -> None:
    """
    Data is multidimensional list, where each inner list corresponds
    to a column.
    """
    extra_space = 3
    data = np.array(data)
    col_widths = [
        int(
            _val_digits(np.amax(np.absolute(data[col])))
            + prec[col]
            + 2
            + extra_space
        )
        for col in range(len(col_names))
    ]
    for i, col in enumerate(col_names):
        out_file.write("{:{width}}".format(col, width=col_widths[i]))
    out_file.write("\n")

    data = data.T
    for row in data:
        for i, val in enumerate(row):
            out_file.write(
                "{:<{width}.{prec}f}".format(
                    val, width=col_widths[i], prec=prec[i]
                )
            )
        out_file.write("\n")


def _val_digits(val: float) -> int:
    """
    Compute the number of decimal digits needed to display the
    integral portion of a value.
    """
    # assume negative for simplicity
    extra_digits = 2

    if val < 10:
        return extra_digits + 1

    return int(np.log10(val)) + extra_digits


def array_index(val, arr) -> int:
    """
    Return the index of the closest array value to a given value.

    :param val: The value for which the closest index is desired.
    :param arr: The array from which the index is computed.

    :returns: The array index whose corresponding value is nearest the
              given value.
    """
    lbound_idx = bisect_left(arr, val)
    if lbound_idx == len(arr):
        return len(arr) - 1
    lbound = arr[lbound_idx]
    if lbound_idx == len(arr):
        return lbound_idx

    ubound_idx = lbound_idx + 1
    ubound = arr[ubound_idx]

    if val - lbound < ubound - val:
        return lbound_idx
    else:
        return ubound_idx


def sort_table_by_col(arr: np.array, col: int = 0):
    """
    Sort a 2D numpy array in ascending order by column index.
    """
    return arr[np.argsort(arr[:, col])]


def table_insertion_idx(val, arr: np.array, col: int = 0):
    """
    Find the insertion index of a value for a sorted 2D numpy array.
    """
    return np.searchsorted(arr[:, col], val)


def interp_lin(xval, xlow, xhigh, ylow, yhigh):
    """
    Get the linear-interpolated y-value for a given x-value between x
    bounds.

    :param xval: The x-value for which you want the y-value.
    :param xlow: The lower-bound x-value.
    :param xhigh: The upper-bound x-value.
    :param ylow: The lower-bound y-value.
    :param yhigh: The upper-bound y-value.
    """
    if xval < xlow or xval > xhigh:
        raise ValueError("xval must be between xlow and xhigh")

    dy = (yhigh - ylow) / (xhigh - xlow)
    dx = xval - xlow
    return ylow + (dy * dx)


def table_interp_val(
    arr: np.array, target_col, sel_val, sel_col: int = 0, permit_outside=False
):
    """
    Get the interpolated column value in a table.

    :param arr: The sorted 2D numpy array.
    :param target_col: Column corresponding to the desired return
        value.
    :param sel_val: Value of the selection column for the desired
        target column.
    :param sel_col: Column index of the selection column.
    :param permit_outside: If True, return lower or upper bound value
        if sel_val is outside table bounds.
    """
    arr = np.array(arr)
    if permit_outside:
        if sel_val < arr[0][sel_col]:
            return arr[0][target_col]
        if sel_val > arr[-1][sel_col]:
            return arr[-1][target_col]

    if sel_val == arr[0][sel_col]:
        return arr[0][target_col]
    if sel_val == arr[-1][sel_col]:
        return arr[-1][target_col]

    ins_idx = table_insertion_idx(sel_val, arr, sel_col)
    xlow = arr[ins_idx - 1][sel_col]
    xhigh = arr[ins_idx][sel_col]
    ylow = arr[ins_idx - 1][target_col]
    yhigh = arr[ins_idx][target_col]

    return interp_lin(sel_val, xlow, xhigh, ylow, yhigh)


def max_priority() -> int:
    """
    Priority that won't be overriden.

    :returns: highest priority.
    """
    return 999


def get_unit(csx: ContinuousStructure) -> float:
    """
    """
    return csx.GetGrid().GetDeltaUnit()


def apply_transform(
    prim: CSPrimitives, transform: CSTransform = None, replace: bool = False
) -> None:
    """
    Apply a transformation to a primitive.  This allows you to pass
    around and apply CSTransform objects directly, rather than having
    to set all transforms via prim.AddTransform.

    :param prim: The primitive to which the transform should be
        applied.
    :param transform: The transformation to apply.
    :param replace: Replace any transforms previously applied to the
        primitive.  Defaults to False, meaning any transforms applied
        here will be applied on top of any existing transforms already
        applied.
    """
    if transform is not None:
        tr = prim.GetTransform()
        concatenate = not replace
        tr.SetMatrix(transform.GetMatrix(), concatenate)


def append_transform(tr1: CSTransform, tr2: CSTransform) -> CSTransform:
    """
    Append two transforms.

    :param tr1: First transform.
    :param tr2: Transform to append to first transform.  The first
        transform will be applied to the primitive first, followed by
        this transform.

    :returns: Combined transform.
    """
    if tr1 is None and tr2 is None:
        return None

    tr_ret = CSTransform()
    if tr1 is not None:
        tr_ret.SetMatrix(tr1.GetMatrix(), True)
    if tr2 is not None:
        tr_ret.SetMatrix(tr2.GetMatrix(), True)
    return tr_ret


def mil_to_mm(mil_val: float) -> float:
    """
    """
    return mil_val * 0.0254


def mm_to_mil(mm_val: float) -> float:
    """
    """
    return mm_val / 0.0254



from typing import List
from pyems.structure import Structure
from pyems.coordinate import Coordinate2


def module_header(name: str) -> str:
    """
    """
    return "(module " + name + " (layer F.Cu) (clearance 0) (attr virtual)\n"


def polygon(poly: List[Coordinate2]) -> str:
    """
    """
    poly_str = "  (fp_poly (pts "
    for coord in poly:
        poly_str += (
            "(xy "
            + "{:.6f}".format(coord.x)
            + " "
            + "{:.6f}".format(-coord.y)
            + ") "
        )

    poly_str += ") (layer F.Cu) (width 0))\n"
    return poly_str


def write_footprint(structure: Structure, name: str, fpath: str):
    """
    Write a Structure as a Kicad footprint.

    :param structure: Structure from which to create a footprint.
    :param name: Footprint name.
    :param fpath: String denoted the path where the footprint file
        should be saved.
    """
    if structure.polygons is None:
        raise RuntimeError("Unable to write footprint for structure.")

    with open(fpath, "w") as fout:
        fout.write(module_header(name))
        for poly in structure.polygons:
            fout.write(polygon(poly))
        fout.write(")\n")



from tempfile import mkdtemp
from enum import Enum
import os
import subprocess
from pyems.simulation import Simulation
from pyems.coordinate import Box3


class DumpType(Enum):
    """
    Field (and other) dump types.
    """

    efield_time = 0
    hfield_time = 1
    current_time = 2
    current_density_time = 3
    efield_frequency = 10
    hfield_frequency = 11
    current_frequency = 12
    current_density_frequency = 13
    local_sar_frequency = 20
    average_sar_frequency_1g = 21
    average_sar_frequency_10g = 22
    raw_data = 29


class FieldDump:
    """
    """

    unique_index = 0

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        dump_type: DumpType = DumpType.efield_time,
        dir_path: str = "dump",
    ):
        """
        :param dir_path: Directory where field dump data is stored.
            The directory is interpreted as being relative to the
            simulation directory.  Therefore, the default will place
            the field dumps within a 'fields' subdirectory of the
            simulation directory.  If left as None, a system temporary
            directory will be used.
        """
        self._sim = sim
        self._box = box
        self._dump_type = dump_type
        self._index = self._get_inc_ctr()
        if dir_path is None:
            dir_path = mkdtemp()
        else:
            dir_path = os.path.abspath(
                os.path.join(
                    self._sim.sim_dir, dir_path + "_" + str(self._index)
                )
            )
            if not os.path.exists(dir_path):
                os.mkdir(dir_path)
        self._dir_path = dir_path

        dump = self._sim.csx.AddDump(
            os.path.join(self._dir_path, "Et_"),
            dump_type=self._dump_type.value,
            file_type=0,
        )
        dump.AddBox(start=self.box.start(), stop=self.box.stop())
        self._sim.add_field_dump(self)

    @property
    def sim(self) -> Simulation:
        """
        """
        return self._sim

    @property
    def box(self) -> Box3:
        """
        """
        return self._box

    @property
    def dump_type(self) -> int:
        """
        """
        return self._dump_type

    def view(self):
        """
        """
        subprocess.run(
            [
                "paraview",
                "--data={}".format(os.path.join(self._dir_path, "Et__..vtr")),
            ]
        )

    @classmethod
    def _get_ctr(cls):
        """
        Retrieve unique counter.
        """
        return cls.unique_index

    @classmethod
    def _inc_ctr(cls):
        """
        Increment unique counter.
        """
        cls.unique_index += 1

    @classmethod
    def _get_inc_ctr(cls):
        """
        Retrieve and increment unique counter.
        """
        ctr = cls._get_ctr()
        cls._inc_ctr()
        return ctr



from typing import List
import numpy as np
from openEMS.ports import UI_data
from pyems.simulation import Simulation
from pyems.mesh import Mesh
from pyems.coordinate import Box3, box_overlap, Axis


# TODO self.csx_box is messy. Should instead wrap CSPrimitives and
# member primitive should have a box.
class Probe:
    """
    """

    unique_ctr = 0

    def __init__(
        self,
        sim: Simulation,
        box: Box3,
        p_type: int = 0,
        normal_axis: Axis = None,
        weight: float = 1,
        mode_function: List = None,
    ):
        """
        """
        self._sim = sim
        self._box = box
        self.p_type = p_type
        self._normal_axis = normal_axis
        self.weight = weight
        self.mode_function = mode_function
        self.name = self._probe_name_prefix() + "t_" + str(self._get_ctr())
        self._inc_ctr()
        self.freq = None
        self.time = None
        self.t_data = None
        self.f_data = None
        self.csx_box = None

        self._set_probe()

    @property
    def sim(self) -> Simulation:
        """
        """
        return self._sim

    @property
    def box(self) -> Box3:
        """
        """
        return self._box

    def pml_overlap(self) -> bool:
        """
        """
        pml_boxes = self.sim.mesh.pml_boxes()
        for pml_box in pml_boxes:
            if box_overlap(self.box, pml_box):
                return True
        return False

    def _set_probe(self) -> None:
        """
        """
        self.csx_probe = self.sim.csx.AddProbe(
            name=self.name, p_type=self.p_type
        )
        self.csx_probe.SetWeighting(self.weight)

        if self._normal_axis is not None:
            self.csx_probe.SetNormalDir(self._normal_axis.axis)

        if self.mode_function is not None:
            self.csx_probe.SetModeFunction(self.mode_function)

        self.csx_box = self.csx_probe.AddBox(
            start=self.box.start(), stop=self.box.stop()
        )

    def snap_to_mesh(self, mesh) -> None:
        """
        Align probe with the provided mesh.  It is necessary to call
        this function in order to get correct simulation results.

        :param mesh: Mesh object.
        """
        for dim in [0, 1, 2]:
            self._snap_dim(mesh, dim)

    def _snap_dim(self, mesh: Mesh, dim: int) -> None:
        """
        Align probe to mesh for a given dimension.  This function will
        only have an effect when the provided dimension has zero size.

        :param mesh: Mesh object.
        :param dim: Dimension.  0, 1, 2 for x, y, z.
        """
        if self.box.min_corner[dim] == self.box.max_corner[dim]:
            start = self.csx_box.GetStart()
            stop = self.csx_box.GetStop()
            _, pos = mesh.nearest_mesh_line(dim, start[dim])
            start[dim] = pos
            stop[dim] = pos
            self.csx_box.SetStart(start)
            self.csx_box.SetStop(stop)

    def read(self, signal_type="pulse"):
        """
        Read data recorded from the simulation and generate the time-
        and frequency-series data.
        """
        freq = self.sim.freq
        sim_dir = self.sim.sim_dir
        self.freq = freq
        data = UI_data([self.name], sim_dir, freq, signal_type)
        self.time = data.ui_time[0]
        self.t_data = data.ui_val[0]
        self.f_data = data.ui_f_val[0]

    def get_freq_data(self) -> np.array:
        """
        Get probe frequency data.

        :returns: 2D numpy array where each inner array is a
                  frequency, value pair.  The result is sorted by
                  ascending frequency.
        """
        if not self._data_readp():
            raise ValueError("Must call read() before retreiving data.")
        else:
            return np.array([self.freq, self.f_data])

    def get_time_data(self):
        """
        Get probe time data.

        :returns: 2D numpy array where each inner array is a
                  time, value pair.  The result is sorted by
                  ascending time.
        """
        if not self._data_readp():
            raise ValueError("Must call read() before retreiving data.")
        else:
            return np.array([self.time, self.t_data]).T

    def _data_readp(self) -> bool:
        if self.freq is not None:
            return True
        else:
            return False

    @classmethod
    def _inc_ctr(cls):
        cls.unique_ctr += 1

    @classmethod
    def _get_ctr(cls):
        return cls.unique_ctr

    def _probe_name_prefix(self):
        if self.p_type == 0:
            return "v"
        elif self.p_type == 1:
            return "i"
        elif self.p_type == 2:
            return "e"
        elif self.p_type == 3:
            return "h"
        elif self.p_type == 10:
            return "wv"
        elif self.p_type == 11:
            return "wi"
        else:
            raise ValueError("invalid p_type")



from typing import List, Tuple
import numpy as np
from pyems.utilities import sort_table_by_col, table_interp_val
from pyems.physical_constant import EPS0


def loss_to_kappa(loss: float, freq: float, epsr: float) -> float:
    """
    """
    return loss * EPS0 * epsr * 2 * np.pi * freq


class Dielectric:
    """
    """

    def __init__(
        self, epsr: List[Tuple[float, float]], loss: List[Tuple[float, float]]
    ):
        """
        :param epsr: Dielectric constant.  List of frequency (Hz) and
            associated dielectric constant.
        :param loss: Loss tangent.  List of frequency (Hz) and
            associated loss tangent.
        """
        self._epsr = sort_table_by_col(np.array(epsr), col=0)
        self._loss = sort_table_by_col(np.array(loss), col=0)
        self._kappa = self._kappa()

    def epsr_at_freq(self, freq: float) -> float:
        """
        Approximate the dielectric at a given frequency given the
        provided epsr values.

        :param freq: frequency of interest (Hz)

        :returns: dielectric constant
        """
        return float(table_interp_val(self._epsr, 1, freq, 0, True))

    def kappa_at_freq(self, freq: float) -> float:
        """
        Approximate the dielectric conductivity at a given frequency
        given the provided epsr and loss values.
        """
        return float(table_interp_val(self._kappa, 1, freq, 0, True))

    def _kappa(self) -> List[Tuple[float, float]]:
        """
        """
        kappas = []
        for freq, loss in self._loss:
            kappas.append(
                (freq, loss_to_kappa(loss, freq, self.epsr_at_freq(freq)))
            )
        return kappas


common_dielectrics = {
    "FR408": Dielectric(
        epsr=[
            (100e6, 3.72),
            (1e9, 3.69),
            (2e9, 3.68),
            (5e9, 3.64),
            (10e9, 3.65),
        ],
        loss=[
            (100e6, 0.0072),
            (1e9, 0.0091),
            (2e9, 0.0092),
            (5e9, 0.0098),
            (10e9, 0.0095),
        ],
    ),
    "PTFE": Dielectric(epsr=[(1, 2.06)], loss=[(1, 0.0002)]),
}



from multiprocessing import Pool
import numpy as np
import scipy.optimize
from scipy.optimize import curve_fit
from scipy.special import polygamma
from pyems.physical_constant import C0, MUE0, Z0
from pyems.utilities import print_table


def wheeler_z0(w: float, t: float, er: float, h: float) -> float:
    """
    Calculate the microstrip characteristic impedance for a given
    width using Wheeler's equation.  Wheeler's equation can be found
    at:

    https://en.wikipedia.org/wiki/Microstrip#Characteristic_impedance

    :param w: microstrip trace width (m)
    :param t: trace thickness (m)
    :param er: substrate relative permittivity
    :param h: substrate height (thickness) (m)

    :returns: characteristic impedance
    """
    weff = w + (
        (t * ((1 + (1 / er)) / (2 * np.pi)))
        * np.log(
            (4 * np.e)
            / (
                np.sqrt(
                    ((t / h) ** 2)
                    + (((1 / np.pi) * ((1 / ((w / t) + (11 / 10))))) ** 2)
                )
            )
        )
    )
    tmp1 = 4 * h / weff
    tmp2 = (14 + (8 / er)) / 11
    zm = (Z0 / (2 * np.pi * np.sqrt(2 * (1 + er)))) * np.log(
        1
        + (
            tmp1
            * (
                (tmp2 * tmp1)
                + (
                    np.sqrt(
                        (
                            (tmp2 * tmp1) ** 2
                            + ((np.pi ** 2) * ((1 + (1 / er)) / 2))
                        )
                    )
                )
            )
        )
    )
    return zm


def wheeler_z0_width(
    z0: float,
    t: float,
    er: float,
    h: float,
    tol: float = 0.01,
    guess: float = 0.3,
) -> float:
    """
    Calculate the microstrip width for a given characteristic
    impedance using Wheeler's formula.

    :param z0: characteristic impedance (ohm)
    :param t: trace thickness (m)
    :param er: substrate relative permittivity
    :param h: substrate height (thickness) (m)
    :param tol: acceptable impedance tolerance (ohm)
    :param guess: an initial guess for the width (m).  This can
        improve convergence time when the approximate width is known.

    :returns: trace width (m)
    """
    width = guess
    zm = wheeler_z0(w=width, t=t, er=er, h=h)
    wlow = width / 10
    zlow = wheeler_z0(w=wlow, t=t, er=er, h=h)
    # inverse relation between width and z0
    while zlow < z0:
        wlow /= 10
        zlow = wheeler_z0(w=wlow, t=t, er=er, h=h)

    whigh = width * 10
    zhigh = wheeler_z0(w=whigh, t=t, er=er, h=h)
    while zhigh > z0:
        whigh *= 10
        zhigh = wheeler_z0(w=whigh, t=t, er=er, h=h)

    while np.absolute(zm - z0) > tol:
        if zm > z0:
            m = (zhigh - zm) / (whigh - width)
            wlow = width
            zlow = zm
        else:
            m = (zm - zlow) / (width - wlow)
            whigh = width
            zhigh = zm

        # use linear interpolation to update guess
        width = width + ((z0 - zm) / m)
        zm = wheeler_z0(w=width, t=t, er=er, h=h)

    return width


def pozar_z0(
    trace_width: float, substrate_height: float, substrate_dielectric: float
) -> float:
    """
    Calculate the characteristic impedance of a microstrip
    transmission line using the approximation in Pozar 4e p.148.
    """
    eeff = microstrip_effective_dielectric(
        substrate_dielectric, substrate_height, trace_width
    )
    wh_ratio = trace_width / substrate_height
    if wh_ratio <= 1:
        return (60 / np.sqrt(eeff)) * np.log((8 / wh_ratio) + (wh_ratio / 4))
    else:
        return (
            120
            * np.pi
            / (
                np.sqrt(eeff)
                * (wh_ratio + 1.393 + 0.667 * np.log(wh_ratio + 1.444))
            )
        )


def pozar_z0_width(
    z0: float, substrate_height: float, substrate_dielectric: float
) -> float:
    """
    Calculate the microstrip trace width for a desired characteristic
    impedance using the approximation in Pozar 4e p.148.
    """
    a = ((z0 / 60) * np.sqrt((substrate_dielectric + 1) / 2)) + (
        ((substrate_dielectric - 1) / (substrate_dielectric + 1))
        * (0.23 + (0.11 / substrate_dielectric))
    )
    b = 377 * np.pi / (2 * z0 * np.sqrt(substrate_dielectric))
    ratio1 = 8 * np.exp(a) / (np.exp(2 * a) - 2)
    ratio2 = (2 / np.pi) * (
        b
        - 1
        - np.log(2 * b - 1)
        + ((substrate_dielectric - 1) / (2 * substrate_dielectric))
        * (np.log(b - 1) + 0.39 - 0.61 / substrate_dielectric)
    )
    if ratio1 < 2 and ratio2 < 2:
        return ratio1 * substrate_height
    elif ratio1 > 2 and ratio2 > 2:
        return ratio2 * substrate_height
    else:
        raise RuntimeError("Conflicting ratios.")


def miter(trace_width: float, substrate_height: float) -> float:
    """
    Compute the optimal miter length using the Douville and James
    equation.
    """
    if trace_width / substrate_height < 0.25:
        raise ValueError(
            "Ratio of trace width to height must be at least 0.25."
        )
    d = trace_width * np.sqrt(2)
    x = d * (
        0.52 + (0.65 * np.exp(-(27 / 20) * trace_width / substrate_height))
    )
    return x


def coax_core_diameter(
    outer_diameter: float, permittivity: float, impedance: float = 50
) -> float:
    """
    Approximate coaxial cable core diameter for a given target
    characteristic impedance and outer diameter.

    :returns: Inner core diameter.  The units will match those of the
              provided outer diameter.
    """
    return outer_diameter / np.power(
        10, impedance * np.sqrt(permittivity) / 138
    )


def microstrip_effective_dielectric(
    substrate_dielectric: float, substrate_height: float, trace_width: float
) -> float:
    """
    Compute the effective dielectric for a microstrip trace.  This
    represents the dielectric constant for a homogenous medium that
    replaces the air and substrate.  See Pozar 4e p.148 and Wadell
    p.94 for details.

    :param substrate_dielectric: Dielectric constant of the PCB
        substrate.
    :param substrate_height: Distance between the backing ground plane
        and microstrip trace.
    :param trace_width: Microstrip trace width.  Any units can be used
        for substrate_height and trace_width as long as they are
        consistent.
    """
    wh_ratio = trace_width / substrate_height
    common_factor = 1 / np.sqrt(1 + (12 / wh_ratio))

    if wh_ratio < 1:
        factor = common_factor + (0.04 * ((1 - wh_ratio) ^ 2))
    else:
        factor = common_factor

    return ((substrate_dielectric + 1) / 2) + (
        ((substrate_dielectric - 1) / 2) * factor
    )


def phase_shift_length(
    phase_shift: float, dielectric: float, frequency: float
) -> float:
    """
    Compute the length (in mm) for a signal to undergo a given phase
    shift.  When computing this value for a transmission line not
    surrounded by a homogenous medium (e.g. a microstrip trace), make
    sure to use the effective dielectric.

    :param phase_shift: Phase shift in degrees.
    :param dielectric: Dielectric or effective dielectric constant.
    :param frequency: Signal frequency.
    """
    rad = phase_shift * np.pi / 180
    vac_lambda = 2 * np.pi * frequency / speed_of_light(unit=1e-3)
    return rad / (np.sqrt(dielectric) * vac_lambda)


def skin_depth(
    frequency: float, resistivity: float = 1.68e-8, rel_permeability: float = 1
) -> float:
    """
    Compute the skin depth for a conductor at a given frequency.  The
    default values are for copper.
    """
    return np.sqrt(
        2 * resistivity / (2 * np.pi * frequency * rel_permeability * MUE0)
    )


def speed_of_light(unit: float) -> float:
    """
    """
    return C0 / unit


def wavelength(freq: np.array, unit: float) -> np.array:
    """
    Calculate the wavelength for a given frequency of light.  This
    presently assumes that the light is travelling through a vacuum.
    """
    return speed_of_light(unit) / freq


def wavenumber(freq: np.array, unit: float) -> np.array:
    """
    Calculate the wavenumber for a given frequency of light.  Assumes
    light is travelling through a vacuum.
    """
    return np.array(2 * np.pi / wavelength(freq, unit))


def sweep(func, params, processes: int = 5):
    """
    :param func: Function object taking one item in the `params` list.
    :param params: List of parameter values to sweep over.
    :param processes: Max number of parallel processes.
    """
    pool = Pool(processes=processes)
    ret_vals = list(pool.map(func, params))
    return ret_vals


def optimize_parameter(
    func, start, step, tol, max_steps, display_progress=False
):
    """
    Compute the lowest-cost value of a parameter that still produces
    accurate results.
    """
    res1 = func(start)
    n = len(res1)
    res_matrix = np.zeros((max_steps, n), dtype=np.clongdouble)
    res_matrix[0] = np.array(res1)
    # compute the root mean square differences from the previous results array.
    rms = np.zeros((max_steps - 1,))
    # If the first few RMS values don't show a clear trend, ignore
    # them, since this may have been the result of poor start value
    # choice.
    rms_trend_down = False
    rms_valid_index = 0
    i = 1
    orig_start = start
    start += step

    while i < max_steps:
        res_matrix[i] = np.array(func(start))
        diff = np.subtract(res_matrix[i], res_matrix[i - 1])
        rms[i - 1] = np.sqrt(
            np.sum(np.real(np.multiply(diff, np.conj(diff)))) / n
        )
        if display_progress:
            num_steps = int(np.round((start - orig_start) / step) + 1)
            print_table(
                np.abs(res_matrix[: i + 1]),
                [
                    "{:.4f}".format(val)
                    for val in np.linspace(orig_start, start, num=num_steps)
                ],
                [4 for _ in range(num_steps)],
            )
            print("parameter: {}".format(start))
            print("RMS: {:.10f}".format(rms[i - 1]))

        if not rms_trend_down and i > 1 and rms[i - 1] > rms[i - 2]:
            rms_valid_index = i - 1

        if i - rms_valid_index > 2:
            if (
                not rms_trend_down
                and rms[i - 1] < rms[i - 2]
                and rms[i - 2] < rms[i - 3]
            ):
                rms_trend_down = True

            valid_start = orig_start + (rms_valid_index * step)
            num_valid_steps = int(
                np.round((start - step - valid_start) / step) + 1
            )
            try:
                print(
                    np.linspace(valid_start, start - step, num=num_valid_steps)
                )
                print(rms[rms_valid_index:i])
                fit = curve_fit(
                    rms_fit,
                    np.linspace(
                        valid_start, start - step, num=num_valid_steps
                    ),
                    rms[rms_valid_index:i],
                )
                a = fit[0][0]
                b = fit[0][1]
                error_estimate = rms_remaining_sum(a, b, start + 1)

                if display_progress:
                    print("Error estimate: {:.10f}".format(error_estimate))

                if error_estimate < tol:
                    return start
            except RuntimeError:
                # If curve fitting fails, ignore the oldest RMS value
                # and try again on the next iteration.
                print(
                    "Failed to fit curve to RMS values, ignoring oldest value."
                )
                rms_valid_index += 1

        i += 1
        start += step

    raise RuntimeError(
        "Failed to optimize parameter. Consider increasing "
        "the tolerance or max number of steps."
    )


def rms_fit(x, a, b):
    """
    """
    return np.divide(a, np.power(np.subtract(x, b), 2))


def rms_remaining_sum(a, b, c):
    """
    Computes the sum: sum(a/(x-b)^2, x=c, oo)
    """
    return a * polygamma(1, c - b)


def minimize(func, initial, tol, bounds=None):
    """
    Thin wrapper around scipy.optimize.minimize.

    I've added this so that you can call minimize directly from pyems
    and to make the calling syntax slightly easier (though less
    flexible).

    :param func: Function object that takes 1 or more arguments and
        returns a result to be minimized.
    :param initial: Initial values passed to `func`.  This must be a
        single value, or an array of values if `func` takes more than
        1 argument.
    :param bounds: Constrains parameters to a range of values.  This
        is a tuple of (min, max) for each parameter.  Use None for no
        bounds.  The default for each parameter is no bounds.  A
        single value of None can be used to not place boundaries on
        any parameters.
    :param tol: Result tolerance.

    :returns: Array of function arguments that minimize the function,
              or a single value if `func` only takes 1 argument.

    Example invocation:
    res = minimize(func=func, initial=[1.2], tol=1e-2, bounds=[(0, None)])
    """
    if bounds is None:
        if type(initial) is list:
            num_params = len(initial)
        else:
            num_params = 1
        bounds = []
        for i in num_params:
            bounds.append((None, None))
        bounds = tuple(bounds)

    res = scipy.optimize.minimize(
        func,
        np.array(initial),
        method="L-BFGS-B",
        tol=tol,
        bounds=bounds,
        options={"disp": True},
    )
    if not res.success:
        raise RuntimeError(
            "Minimization failed. See scipy.optimize.minimize "
            "for other options."
        )

    if len(res.x) == 1:
        return res.x[0]
    else:
        return res.x



from typing import List
import numpy as np
from pyems.material import Dielectric, common_dielectrics


# TODO this might need to be adapted in the future to support a
# different dielectric for each substrate layer.
class PCBProperties:
    """
    A PCB structure and material properties for use in an openems simulation.
    """

    def __init__(
        self,
        substrate: Dielectric,
        copper_thickness: List[float],
        substrate_thickness: List[float],
        metal_conductivity: float,
        via_plating_thickness: float,
    ):
        """
        :param substrate: Substrate dielectric material.
        :param copper_thickness: thickness of each conductive layer
            (in m).  Again proceeds from top to bottom layer.
        :param substrate_thickness: separations (in m) between
            adjacent copper layers.  A list where the first value is
            the separation between the top layer and second layer,
            etc.  This is equivalently the substrate thickness.
        :param metal_conductivity: Metal layer conductivity in S/m.
        :param via_plating_thickness: Thickness of the via plating (in
            m).
        """
        self._substrate = substrate
        self.copper_thick = copper_thickness
        self.substrate_thick = substrate_thickness
        self.metal_kappa = metal_conductivity
        self.via_plating_thick = via_plating_thickness

    @property
    def substrate(self) -> Dielectric:
        """
        """
        return self._substrate

    def copper_thickness(self, index: int, unit: float = 1) -> float:
        """
        Get the thickness of a copper layer.

        :param index: The index of the copper layer for which the
            thickness should be given.  The index ranges from 0 (for
            the top layer) to 1 less than the number of layers (bottom
            index).
        """
        if index < 0 or index >= len(self.copper_thick):
            raise ValueError("Invalid index provided.")

        return self.copper_thick[index] / unit

    def substrate_thickness(self, index: int, unit: float = 1) -> float:
        """
        Get the thickness of a substrate layer.

        :param index: The index of the substrate layer for which the
            thickness should be given.  The index ranges from 0 (for
            the substrate layer between the top and next copper
            layers) to 1 less than the number of substrate layers.
        """
        if index < 0 or index >= len(self.substrate_thick):
            raise ValueError("Invalid index provided.")

        return self.substrate_thick[index] / unit

    def via_plating_thickness(self, unit: float = 1) -> float:
        """
        """
        return self.via_plating_thick / unit

    def metal_conductivity(self) -> float:
        """
        """
        return self.metal_kappa

    def num_copper_layers(self) -> int:
        """
        Number of copper layers.

        :returns: The number of copper layers.
        """
        return len(self.copper_thick)

    def num_layers(self) -> int:
        """
        Total number of PCB layers.

        :returns: The total number of PCB layers, including all copper
                  and substrate layers.
        """
        return len(self.copper_thick) + len(self.substrate_thick)

    def copper_layer_dist(
        self,
        layer: int,
        unit: float = 1,
        ref_layer: int = 0,
        zero_thickness: bool = True,
    ) -> float:
        """
        Get the distance of a copper layer from some other layer.

        :param layer: The index of the copper layer for which the
            separation from the reference layer should be calculated.
            Indexed from 0 to 1 less than the number of copper layers.
        :param unit: Dimensional unit.  Leave the default for meters.
        :param ref_layer: The reference layer from which to compute
            the distance.  This defaults to the top layer.
        :param zero_thickness: Assume all copper layers contribute
            zero thickness.  This is the default since OpenEMS yields
            faster and more accurate simulations if zero-thickness
            ConductingSheet's are used for thin metals than if
            non-zero thickness PEC's are used.  If you're using
            non-zero thickness metals, set this to false.  When false,
            we measure to the top layer of the target layer from the
            top of the reference layer.
        """
        num_copper_layers = self.num_copper_layers()
        if (
            layer >= num_copper_layers
            or layer < 0
            or ref_layer >= num_copper_layers
            or ref_layer < 0
        ):
            raise ValueError("Invalid layer or ref_layer index.")

        if layer < ref_layer:
            raise ValueError(
                "layer should have at least as high a reference index as "
                "the reference layer."
            )

        thickness = np.sum(self.substrate_thick[ref_layer:layer])
        if not zero_thickness:
            thickness += np.sum(self.copper_thick[ref_layer:layer])

        return thickness / unit


# trace specs:
# min trace width = 5mil (0.127mm)
# min trace spacing = 5mil (0.127mm)
#
# drill specs:
# min annular ring = 4mil (0.1016mm)
# min drill size = 10mil (0.254mm)
common_pcbs = {
    "oshpark4": PCBProperties(
        substrate=common_dielectrics["FR408"],
        copper_thickness=[0.0356e-3, 0.0178e-3, 0.0178e-3, 0.0356e-3],
        substrate_thickness=[0.1702e-3, 1.1938e-3, 0.1702e-3],
        metal_conductivity=5.8e7,
        via_plating_thickness=0.0254e-3,
    )
}



from typing import List
from copy import deepcopy
import numpy as np


def val_inside(val: float, bound1: float, bound2: float) -> bool:
    """
    """
    min_bound = np.min([bound1, bound2])
    max_bound = np.max([bound1, bound2])
    if val >= min_bound and val <= max_bound:
        return True
    return False


class Coordinate2:
    """
    """

    def __init__(self, x: float, y: float):
        """
        """
        self._x = x
        self._y = y

    @property
    def x(self) -> float:
        """
        """
        return self._x

    @property
    def y(self) -> float:
        """
        """
        return self._y

    @x.setter
    def x(self, value: float) -> None:
        """
        """
        self._x = value

    @y.setter
    def y(self, value: float) -> None:
        """
        """
        self._y = value

    def __getitem__(self, key):
        """
        """
        return self._int_to_coord(key)

    def __setitem__(self, key, val):
        """
        """
        if key == 0:
            self.x = val
        elif key == 1:
            self.y = val
        else:
            raise ValueError("Invalid index.")

    def _int_to_coord(self, val) -> float:
        """
        """
        if val == 0:
            return self.x
        elif val == 1:
            return self.y
        else:
            raise ValueError("Invalid index.")

    def coordinate_list(self) -> List[float]:
        """
        """
        return [self._x, self._y]


class Coordinate3(Coordinate2):
    """
    """

    def __init__(self, x: float, y: float, z: float):
        """
        """
        self._z = z
        super().__init__(x, y)

    @property
    def z(self):
        """
        """
        return self._z

    @z.setter
    def z(self, value: float):
        """
        """
        self._z = value

    def coordinate_list(self) -> List[float]:
        """
        """
        return [self._x, self._y, self._z]

    def __setitem__(self, key, val):
        """
        """
        if key == 0:
            self.x = val
        elif key == 1:
            self.y = val
        elif key == 2:
            self.z = val
        else:
            raise ValueError("Invalid index.")

    def _int_to_coord(self, val):
        """
        """
        if val == 0:
            return self.x
        elif val == 1:
            return self.y
        elif val == 2:
            return self.z
        else:
            raise ValueError("Invalid index.")


class Axis:
    """
    An abstraction of a coordinate axis.  It may also optionally
    specity a direction for cases where it is significant whether the
    axis points in the positive or negative direction relative to the
    normal coordinate axis.  When a direction is not specified, the
    positive direction is taken.
    """

    def __init__(self, axis, direction: int = 1):
        """
        """
        if type(axis) is str:
            self._axis = self._str_to_int(axis)
        else:
            self._axis = axis

        self._direction = direction
        self._check_direction()

    @property
    def axis(self) -> int:
        """
        """
        return self._axis

    @property
    def direction(self) -> int:
        """
        """
        return self._direction

    def is_positive_direction(self) -> bool:
        """
        """
        return self._direction == 1

    def as_list(self) -> List[int]:
        """
        """
        lst = [0, 0, 0]
        lst[self._axis] = self._direction
        return lst

    def _check_direction(self) -> None:
        """
        """
        if self._direction != 1 and self._direction != -1:
            raise ValueError("Invalid direction. Valid values are +1 and -1.")

    def _str_to_int(self, val: str) -> int:
        """
        """
        lval = val.lower()
        if lval == "x":
            return 0
        elif lval == "y":
            return 1
        elif lval == "z":
            return 2
        else:
            raise ValueError(
                "Invalid string axis value. Valid values are "
                "'x', 'y', and 'z' (case insensitive)."
            )

    def _int_to_str(self, val: int) -> str:
        """
        """
        if val == 0:
            return "x"
        elif val == 1:
            return "y"
        elif val == 2:
            return "z"
        else:
            raise ValueError(
                "Invalid integer axis value. Valid values are 0, 1, and 2."
            )

    def intval(self) -> int:
        """
        """
        return self._axis

    def strval(self) -> str:
        """
        """
        return self._int_to_str(self._axis)


class Box2:
    """
    """

    def __init__(self, min_corner: Coordinate2, max_corner: Coordinate2):
        """
        """
        self._min_corner = min_corner
        self._max_corner = max_corner

    @property
    def min_corner(self) -> Coordinate2:
        """
        """
        return deepcopy(self._min_corner)

    @property
    def max_corner(self) -> Coordinate2:
        """
        """
        return deepcopy(self._max_corner)

    def start(self) -> List[float]:
        """
        List object expected by OpenEMS interface.
        """
        return self.min_corner.coordinate_list()

    def stop(self) -> List[float]:
        """
        List object expected by OpenEMS interface.
        """
        return self.max_corner.coordinate_list()

    def as_list(self) -> List[List[float]]:
        """
        """
        return [self.start(), self.stop()]

    def corners(self) -> List[Coordinate2]:
        """
        """
        # TODO should be a more concise way to do this.
        corner1 = deepcopy(self.min_corner)

        corner2 = deepcopy(self.min_corner)
        corner2.y = self.max_corner.y

        corner3 = deepcopy(self.max_corner)

        corner4 = deepcopy(self.min_corner)
        corner4.x = self.max_corner.x

        return [corner1, corner2, corner3, corner4]

    def center(self) -> Coordinate2:
        """
        """
        return Coordinate2(
            np.average([self.min_corner.x, self.max_corner.x]),
            np.average([self.min_corner.y, self.max_corner.y]),
        )

    def length(self) -> float:
        """
        """
        return np.abs(self.max_corner.x - self.min_corner.x)

    def width(self) -> float:
        """
        """
        return np.abs(self.max_corner.y - self.min_corner.y)

    def negative_direction(self) -> bool:
        """
        """
        return self.max_corner.x < self.min_corner.x

    def has_zero_dim(self) -> bool:
        """
        Return True if at least 1 dimension of the box has zero size.
        """
        if (
            self.max_corner.x - self.min_corner.x == 0
            or self.max_corner.y - self.min_corner.y == 0
        ):
            return True
        return False


class Box3:
    """
    """

    def __init__(self, min_corner: Coordinate3, max_corner: Coordinate3):
        """
        """
        self._min_corner = min_corner
        self._max_corner = max_corner

    def __getitem__(self, key):
        """
        """
        return self._int_to_corner(key)

    def _int_to_corner(self, val: int) -> Coordinate3:
        """
        """
        if val == 0:
            return self.min_corner
        elif val == 1:
            return self.max_corner
        else:
            raise ValueError("Invalid Box3 index.")

    @property
    def min_corner(self) -> Coordinate3:
        """
        """
        return self._min_corner

    @min_corner.setter
    def min_corner(self, val: Coordinate3) -> None:
        """
        """
        self._min_corner = val

    @property
    def max_corner(self) -> Coordinate3:
        """
        """
        return self._max_corner

    @max_corner.setter
    def max_corner(self, val: Coordinate3) -> None:
        """
        """
        self._max_corner = val

    def set_increasing(self) -> None:
        """
        Rearrange min and max coordinates so that all min_corner
        coordinates are less than or equal to all max_corner
        coordinates.
        """
        for dim in range(3):
            if self.min_corner[dim] > self.max_corner[dim]:
                old_max = self.max_corner[dim]
                self.max_corner[dim] = self.min_corner[dim]
                self.min_corner[dim] = old_max

    def corners(self) -> List[Coordinate3]:
        """
        """
        # TODO should be a more concise way to do this.
        corner1 = deepcopy(self.min_corner)

        corner2 = deepcopy(self.min_corner)
        corner2.y = self.max_corner.y

        corner3 = deepcopy(self.min_corner)
        corner3.x = self.max_corner.x

        corner4 = deepcopy(self.max_corner)
        corner4.z = self.min_corner.z

        corner5 = deepcopy(self.min_corner)
        corner5.z = self.max_corner.z

        corner6 = deepcopy(self.max_corner)
        corner6.x = self.min_corner.x

        corner7 = deepcopy(self.max_corner)
        corner7.y = self.min_corner.y

        corner8 = deepcopy(self.max_corner)

        return [
            corner1,
            corner2,
            corner3,
            corner4,
            corner5,
            corner6,
            corner7,
            corner8,
        ]

    def inside(self, point: Coordinate3) -> bool:
        """
        """
        for dim in range(3):
            if not val_inside(
                point[dim], self.min_corner[dim], self.max_corner[dim]
            ):
                return False
        return True

    def start(self) -> List[float]:
        """
        List object expected by OpenEMS interface.
        """
        return self.min_corner.coordinate_list()

    def stop(self) -> List[float]:
        """
        List object expected by OpenEMS interface.
        """
        return self.max_corner.coordinate_list()

    def as_list(self) -> List[List[float]]:
        """
        """
        return [self.start(), self.stop()]

    def center(self) -> Coordinate3:
        """
        """
        return Coordinate3(
            np.average([self.min_corner.x, self.max_corner.x]),
            np.average([self.min_corner.y, self.max_corner.y]),
            np.average([self.min_corner.z, self.max_corner.z]),
        )

    def length(self) -> float:
        """
        """
        return np.abs(self.max_corner.x - self.min_corner.x)

    def width(self) -> float:
        """
        """
        return np.abs(self.max_corner.y - self.min_corner.y)

    def height(self) -> float:
        """
        """
        return np.abs(self.max_corner.z - self.min_corner.z)

    def negative_direction(self) -> bool:
        """
        """
        return self.max_corner.x < self.min_corner.x

    def has_zero_dim(self) -> bool:
        """
        Return True if at least 1 dimension of the box has zero size.
        """
        if (
            self.max_corner.x - self.min_corner.x == 0
            or self.max_corner.y - self.min_corner.y == 0
            or self.max_corner.z - self.min_corner.z == 0
        ):
            return True
        return False


def box_overlap(box1: Box3, box2: Box3) -> bool:
    """
    """
    box1_corners = box1.corners()
    for corner in box1_corners:
        if box2.inside(corner):
            return True

    box2_corners = box2.corners()
    for corner in box2_corners:
        if box1.inside(corner):
            return True

    return False
